#!/usr/bin/env python
# -*- coding: utf-8 -*-
r"""
@DATE: 2024/4/21 16:57
@File: pytest_log.py
@IDE: pycharm
@Description:
    测试swanlog类，只需测试其日志监听功能
"""
import os

import pytest
from freezegun import freeze_time
from nanoid import generate

from swanlab.log import swanlog
from tutils import TEMP_PATH


@pytest.fixture(scope="function", autouse=True)
def before_test_global_swanlog():
    """
    在测试之前清除全局swanlog的状态
    """
    try:
        swanlog.uninstall()
    except RuntimeError:
        pass


class TestSwanLogInstall:
    """
    目前在设计上不希望外界实例化SwanLog，所以不提供实例化测试
    """

    @staticmethod
    def create_console_dir():
        # 创建console文件夹
        console_dir = os.path.join(TEMP_PATH, str(generate()))
        os.mkdir(console_dir)
        return console_dir

    def test_global_install(self):
        swanlog.install()
        assert swanlog.installed is True
        with pytest.raises(RuntimeError) as e:
            swanlog.install()
        assert str(e.value) == "SwanLog has been installed"
        swanlog.uninstall()
        swanlog.install()
        assert swanlog.installed is True

    def test_write_after_uninstall(self):
        console_dir = self.create_console_dir()
        swanlog.install(console_dir)
        swanlog.uninstall()
        # 加一行防止其他问题
        print("\ntest write after uninstall")
        a = generate()
        print(a)
        b = generate()
        print(b)
        files = os.listdir(console_dir)
        assert len(files) == 1
        with open(os.path.join(console_dir, files[0]), "r") as f:
            content = f.readlines()
            assert len(content) == 0

    def test_write_to_file(self):
        console_dir = self.create_console_dir()
        swanlog.install(console_dir)
        # 加一行防止其他问题
        print("\ntest write to file")
        a = generate()
        print(a)
        b = generate()
        print(b)
        files = os.listdir(console_dir)
        assert len(files) == 1
        # 比较最后两行内容
        with open(os.path.join(console_dir, files[0]), "r") as f:
            content = f.readlines()
            assert content[-2] == a + "\n"
            assert content[-1] == b + "\n"

    def test_write_to_file_long_test(self):
        console_dir = self.create_console_dir()
        swanlog.install(console_dir)
        # 加一行防止其他问题
        print("\ntest write to file")
        a = generate(size=501)
        print(a)
        files = os.listdir(console_dir)
        with open(os.path.join(console_dir, files[0]), "r") as f:
            content = f.readlines()
            assert content[-1] == a[:500] + "\n"

    def test_write_logging_to_file(self):
        console_dir = self.create_console_dir()
        swanlog.install(console_dir, log_level="debug")
        # 加一行防止其他问题
        print("\ntest write to file")
        a = generate()
        swanlog.debug(a)
        b = generate()
        swanlog.info(b)
        files = os.listdir(console_dir)
        assert len(files) == 1
        with open(os.path.join(console_dir, files[0]), "r") as f:
            content = f.readlines()
            assert content[-2] == "swanlab: " + a + "\n"
            assert content[-1] == "swanlab: " + b + "\n"

    def test_can_write_logging(self):
        console_dir = self.create_console_dir()
        swanlog.install(console_dir)
        # 加一行防止其他问题
        print("\ntest write to file")
        a = generate()
        swanlog.debug(a)
        b = generate()
        swanlog.info(b)
        files = os.listdir(console_dir)
        assert len(files) == 1
        with open(os.path.join(console_dir, files[0]), "r") as f:
            content = f.readlines()
            assert content[-2] != "swanlab: " + a + "\n"
            assert content[-1] == "swanlab: " + b + "\n"

    def test_write_sharding(self, monkeypatch):
        """
        测试日志文件分片
        """
        console_dir = self.create_console_dir()
        with freeze_time('2020-10-06'):
            swanlog.install(console_dir)
            print("1234")
            assert os.path.exists(os.path.join(console_dir, "2020-10-06.log"))
        with freeze_time('2020-10-07'):
            p = os.path.join(console_dir, "2020-10-07.log")
            assert not os.path.exists(p)
            print("1234")
            assert os.path.exists(p)



# 导出初始化函数和log函数
from .data import (
    login,
    init,
    log,
    finish,
    Audio,
    Image,
    Text,
    Run,
    State,
    get_run,
    get_config,
    config,
    get_url,
    get_project_url,
)
from .env import SwanLabEnv
from .package import get_package_version
from .sync import sync_wandb, sync_tensorboardX, sync_tensorboard_torch

# 设置默认环境变量
SwanLabEnv.set_default()
# 检查当前需要检查的环境变量
SwanLabEnv.check()

__version__ = get_package_version()

__all__ = [
    "login",
    "init",
    "log",
    "finish",
    "Audio",
    "Image",
    "Text",
    "Run",
    "State",
    "get_run",
    "get_config",
    "config",
    "__version__",
]



import json
from typing import cast, Any, Dict
import xgboost as xgb  # type: ignore
from xgboost import Booster
import swanlab


class SwanLabCallback(xgb.callback.TrainingCallback):
    def __init__(self):
        # 如果没有注册过实验
        swanlab.config["FRAMEWORK"] = "xgboost"
        if swanlab.get_run() is None:
            raise RuntimeError("You must call swanlab.init() before SwanLabCallback(). 你必须在SwanLabCallback()之前，调用swanlab.init().")
    
    def update_config(self, config: Dict[str, Any]):
        swanlab.config.update(config)
        
    def before_training(self, model: Booster) -> Booster:
        """Run before training is finished."""
        # Update SwanLab config
        config = model.save_config()
        swanlab.config.update(json.loads(config))

        return model

    def after_training(self, model: Booster) -> Booster:
        """Run after training is finished."""

        # Log the best score and best iteration
        if model.attr("best_score") is not None:
            swanlab.log(
                {
                    "best_score": float(cast(str, model.attr("best_score"))),
                    "best_iteration": int(cast(str, model.attr("best_iteration"))),
                }
            )

        return model

    def after_iteration(self, model: Booster, epoch: int, evals_log: dict) -> bool:
        """Run after each iteration. Return True when training should stop."""
        # Log metrics
        for data, metric in evals_log.items():
            for metric_name, log in metric.items():
                    swanlab.log({f"{data}-{metric_name}": log[-1]})

        swanlab.log({"epoch": epoch})

        return False




import json
import math
from typing import Dict, Optional

from swankit.callback.models import MetricInfo, ColumnInfo, MetricErrorInfo, ColumnClass, SectionType, ColumnConfig
from swankit.core import SwanLabSharedSettings
from swankit.env import create_time

from swanlab.data.modules import DataWrapper, Line
from swanlab.log import swanlog
from .helper import SwanLabRunOperator


class SwanLabExp:
    """
    Class for running experiments
    save keys when running experiments
    """

    def __init__(self, settings: SwanLabSharedSettings, operator: SwanLabRunOperator) -> None:
        """初始化实验

        Parameters
        ----------
        settings : SwanLabSharedSettings
            全局运行时配置
        operator : SwanLabRunOperator
            操作员
        """
        self.settings = settings
        # 当前实验的所有tag数据字段
        self.keys: Dict[str, SwanLabKey] = {}
        # TODO 操作员不传递给实验
        self.__operator = operator

    def __add(
        self,
        key: str,
        name: Optional[str],
        column_class: ColumnClass,
        column_config: Optional[ColumnConfig],
        section_type: SectionType,
        data: DataWrapper,
        step: int = None,
    ) -> MetricInfo:
        """记录一条新的key数据

        Parameters
        ----------
        key : str
            key的云端唯一标识
        name : str
            key的实际名称
        column_class : str
            列类型，CUSTOM为自定义key，SYSTEM为系统key
        section_type : str
            key的组类型
        data : DataWrapper
            包装后的数据，用于数据解析
        step : int, optional
            步数，如果不传则默认当前步数为'已添加数据数量+1'
            在log函数中已经做了处理，此处不需要考虑数值类型等情况
        """
        key_index = f"{column_class}-{key}"
        # 判断tag是否存在，如果不存在则创建tag
        key_obj: SwanLabKey = self.keys.get(key_index, None)

        # ---------------------------------- 包装器解析 ----------------------------------

        if step is not None and not isinstance(step, int):
            swanlog.warning(f"Step {step} is not int, SwanLab will set it automatically.")
            step = None
        if key_obj is None:
            step = 0 if step is None or not isinstance(step, int) else step
        else:
            step = len(key_obj.steps) if step is None else step
            if step in key_obj.steps:
                swanlog.warning(f"Step {step} on key {key} already exists, ignored.")
                return MetricErrorInfo(column_info=key_obj.column_info, error=DataWrapper.create_duplicate_error())
        data.parse(step=step, key=key)

        # ---------------------------------- 图表创建 ----------------------------------

        if key_obj is None:
            num = len(self.keys)
            # 将此tag对象添加到实验列表中
            key_obj = SwanLabKey(key, self.settings)
            self.keys[key_index] = key_obj
            # 新建图表，完成数据格式校验
            column_info = key_obj.create_column(
                key,
                name,
                column_class,
                column_config,
                section_type,
                data,
                num,
            )
            self.warn_type_error(key_index, key)
            # 创建新列，生成回调
            self.__operator.on_column_create(column_info)

        # 检查tag创建时图表是否创建成功，如果失败则也没有写入数据的必要了，直接退出
        if not key_obj.is_chart_valid:
            self.warn_chart_error(key_index, key)
            return MetricErrorInfo(key_obj.column_info, error=key_obj.column_info.error)
        key_info = key_obj.add(data)
        key_info.buffers = data.parse().buffers
        key_info.media_dir = self.settings.media_dir
        return key_info

    def add(
        self,
        data: DataWrapper,
        key: str,
        name: str = None,
        column_class: ColumnClass = 'CUSTOM',
        column_config: Optional[ColumnConfig] = None,
        section_type: SectionType = "PUBLIC",
        step: int = None,
    ) -> MetricInfo:
        """记录一条新的key数据
        Parameters
        ----------
        data : DataWrapper
            包装后的数据，用于数据解析
        key : str
            列的云端唯一标识
        name : str
            列的实际名称, 默认与key相同
        column_class : str, optional
            列的类型
        column_config : Optional[ColumnConfig], optional
            列的额外配置信息
        section_type : str, optional
            key的组类型
        step : int, optional
            步数，如果不传则默认当前步数为'已添加数据数量+1'
            在log函数中已经做了处理，此处不需要考虑数值类型等情况
        """
        m = self.__add(key, name, column_class, column_config, section_type, data, step)
        self.__operator.on_metric_create(m)
        return m

    def warn_type_error(self, key_index: str, key: str):
        """警告类型错误
        执行此方法时需保证key已经存在
        """
        tag_obj = self.keys[key_index]
        class_name = tag_obj.column_info.got
        expected = tag_obj.column_info.expected
        if tag_obj.is_chart_valid:
            return
        if class_name == "list":
            swanlog.error(f"Data type error, key: {key}, there is element of invalid data type in the list.")
        else:
            swanlog.error(f"Data type error, key: {key}, data type: {class_name}, expected: {expected}")

    def warn_chart_error(self, key_index: str, key: str):
        """
        警告图表创建错误
        执行此方法时需保证key已经存在
        """
        tag_obj = self.keys[key_index]
        if tag_obj.is_chart_valid:
            return
        class_name = tag_obj.column_info.got
        if class_name == "list":
            swanlog.warning(
                f"Chart '{key}' creation failed. "
                f"Reason: The data type in list of the key '{key}' is not as expected, please check the data type."
            )
        else:
            swanlog.warning(
                f"Chart '{key}' creation failed. "
                f"Reason: The expected value type for the chart '{key}' is one of int,"
                f"float or BaseType, but the input type is {class_name}."
            )


class SwanLabKey:
    """
    运行时每个tag的配置，用于记录一些信息
    包括每个tag专属的图表配置
    """

    # 每__slice_size个tag数据保存为一个文件
    __slice_size = 1000

    def __init__(self, key: str, settings: SwanLabSharedSettings) -> None:
        """
        初始化tag对象

        Parameters
        ----------
        key : str
            列名称
        settings : SwanLabSharedSettings
            全局运行时配置
        """
        self.key = key
        self.__steps = set()
        """
        此tag已经包含的steps步骤
        """
        self.__settings = settings
        self.__summary = {}
        """数据概要总结"""
        self.__collection = self.__new_metric_collection()
        """当前tag的数据"""
        self.chart = None
        """当前tag的数据类型，如果是BaseType类型，则为BaseType的小写类名，否则为default"""
        self.__column_info = None

    @property
    def sum(self):
        """当前tag的数据总数"""
        return len(self.__steps)

    @property
    def column_info(self) -> Optional[ColumnInfo]:
        """获取当前tag对应的ColumnInfo"""
        return self.__column_info

    @property
    def steps(self):
        """获取当前tag的所有步数"""
        return self.__steps

    @property
    def chart_created(self):
        """判断当前tag是否已经创建了图表"""
        return self.__column_info is not None

    @property
    def is_chart_valid(self) -> bool:
        """判断当前tag对应的自动创建图表是否成功"""
        return self.column_info.error is None

    def add(self, data: DataWrapper) -> MetricInfo:
        """添加一个数据，在内部完成数据类型转换
        如果转换失败，打印警告并退出
        并且添加数据，当前的数据保存是直接保存，后面会改成缓存形式
        进入此函数之前column_info必须已经创建

        Parameters
        ----------
        data : DataType
            待添加的数据

        Returns
        -------
        metric_info
            添加数据的信息，如果ok为False，则返回None
        ok
            是否添加成功，如果当前step已经存在，或者类型转换失败，返回False
        """
        if self.__column_info is None:
            raise ValueError("Column info is None, please create column info first")
        result = data.parse()
        if data.error is not None:
            swanlog.warning(
                f"Log failed. Reason: Data on key '{self.key}' (step {result.step}) cannot be converted ."
                f"It should be {data.error.expected}, but it is {data.error.got}, please check the data type."
            )
            return MetricErrorInfo(column_info=self.__column_info, error=data.error)

        # 如果为Line且为NaN或者INF，不更新summary
        r = result.strings or result.float

        if not data.type == Line or r not in [Line.nan, Line.inf]:
            if self.__summary.get("max") is None or r > self.__summary["max"]:
                self.__summary["max"] = r
                self.__summary["max_step"] = result.step
            if self.__summary.get("min") is None or r < self.__summary["min"]:
                self.__summary["min"] = r
                self.__summary["min_step"] = result.step
        self.__summary["num"] = self.__summary.get("num", 0) + 1
        self.__steps.add(result.step)
        swanlog.debug(f"Add data, key: {self.key}, step: {result.step}, data: {r}")
        if len(self.__collection["data"]) >= self.__slice_size:
            self.__collection = self.__new_metric_collection()

        new_data = self.__new_metric(result.step, r, more=result.more)
        self.__collection["data"].append(new_data)
        epoch = len(self.__steps)
        mu = math.ceil(epoch / self.__slice_size)
        return MetricInfo(
            column_info=self.__column_info,
            metric=json.loads(json.dumps(new_data)),
            metric_summary=json.loads(json.dumps(self.__summary)),
            metric_epoch=epoch,
            metric_step=result.step,
            metric_buffers=result.buffers,
            metric_file_name=str(mu * self.__slice_size) + ".log",
            swanlab_logdir=self.__settings.log_dir,
            swanlab_media_dir=self.__settings.media_dir if result.buffers else None,
        )

    def create_column(
        self,
        key: str,
        name: Optional[str],
        column_class: ColumnClass,
        column_config: Optional[ColumnConfig],
        section_type: SectionType,
        data: DataWrapper,
        num: int,
    ) -> ColumnInfo:
        """
        创建列信息，对当前key的基本信息做一个记录
        :param key: str, key名称
        :param name: str, key的实际名称
        :param column_class: str, key的类型
        :param column_config: ColumnConfig, key的配置
        :param section_type: str, key的组类型
        :param data: DataType, 数据
        :param num: 创建此列之前的列数量
        """
        if self.chart_created:
            raise ValueError(f"Chart {key} has been created, cannot create again.")
        result = data.parse()

        # 如果section_type不为public,则section_name为None
        if section_type != "PUBLIC":
            result.section = None
        # 如果斜杠，则使用斜杠前的部分作为section的名称
        elif "/" in key and key[0] != "/":
            result.section = key.split("/")[0]

        column_info = ColumnInfo(
            key=key,
            kid=str(num),
            name=name,
            cls=column_class,
            config=column_config,
            chart_type=result.chart,
            section_name=result.section,
            section_type=section_type,
            section_sort=None,
            chart_reference=result.reference,
            error=data.error,
        )
        self.chart = result.chart.value
        self.__column_info = column_info
        return column_info

    @staticmethod
    def __new_metric(index, data, more: dict = None) -> dict:
        """创建一个新的data数据，实际上是一个字典，包含一些默认信息

        Parameters
        ----------
        index : int
            步数
        data : DataType
            数据
        more : dict, optional
            更多的数据，如果有的话
        """
        if more is None:
            return {
                "index": int(index),
                "data": data,
                "create_time": create_time(),
            }
        else:
            return {
                "index": int(index),
                "data": data,
                "create_time": create_time(),
                "more": more,
            }

    @staticmethod
    def __new_metric_collection() -> dict:
        """创建一个新的key data数据集合

        Returns
        -------
        dict
            返回一个新的data数据集合
        """
        time = create_time()
        return {
            "create_time": time,
            "update_time": time,
            "data": [],
        }



"""
@author: cunyue
@file: requirements.py
@time: 2024/11/18 15:48
@description: 依赖包信息采集
"""

import subprocess


def get_requirements():
    """获取当前环境依赖"""
    try:
        # 运行pip命令获取当前环境下的环境目录
        result = subprocess.run(["pip", "list", "--format=freeze"], stdout=subprocess.PIPE, text=True)
        # 检查命令是否成功运行
        if result.returncode == 0:
            return result.stdout
        else:
            return None
    except Exception:  # noqa
        return None



import wandb
import random
import swanlab

swanlab.sync_wandb(wandb_run=False)

wandb.init(
  project="test",
  config={"a": 1, "b": 2},
  name="test",
  notes="test_wandb_sync",
  )

wandb.config.update({"c": 3, "d": 4})
print(swanlab.config.get("c"))

epochs = 10
offset = random.random() / 5
for epoch in range(2, epochs):
  acc = 1 - 2 ** -epoch - random.random() / epoch - offset
  loss = 2 ** -epoch + random.random() / epoch + offset

  wandb.log({"acc": acc, "loss": loss})


#!/usr/bin/env python
# -*- coding: utf-8 -*-
r"""
@DATE: 2024/9/20 14:35
@File: namer.py
@IDE: pycharm
@Description:
    命名器、取色器
"""
import random
from typing import Tuple, Optional

prefix_list = [
    "swan-",  # 天鹅
    "rat-",  # 鼠
    "ox-",  # 牛
    "tiger-",  # 虎
    "rabbit-",  # 兔
    "dragon-",  # 龙
    "snake-",  # 蛇
    "horse-",  # 马
    "goat-",  # 羊
    "monkey-",  # 猴
    "rooster-",  # 鸡
    "dog-",  # 狗
    "pig-",  # 猪
    "cat-",  # 猫
    "elephant-",  # 象
    "penguin-",  # 企鹅
    "kangaroo-",  # 袋鼠
    "monkey-",  # 猴
    "panda-",  # 熊猫
    "tiger-",  # 虎
    "lion-",  # 狮
    "zebra-",  # 斑马
]


def generate_name(index: Optional[int] = None) -> str:
    """
    生成名称
    :param index: 生成名称的索引，约定为历史实验数，可以为None
    :return: 生成的名称
    """
    if index is None:
        prefix = random.choice(prefix_list)
        return prefix
    else:
        prefix = prefix_list[index % len(prefix_list)]
        return prefix + str(index + 1)


light_colors = [
    "#528d59",  # 绿色
    "#587ad2",  # 蓝色
    "#c24d46",  # 红色
    "#9cbe5d",  # 青绿色
    "#6ebad3",  # 天蓝色
    "#dfb142",  # 橙色
    "#6d4ba4",  # 紫色
    "#8cc5b7",  # 淡青绿色
    "#892d58",  # 紫红色
    "#40877c",  # 深青绿色
    "#d0703c",  # 深橙色
    "#d47694",  # 粉红色
    "#e3b292",  # 淡橙色
    "#b15fbb",  # 浅紫红色
    "#905f4a",  # 棕色
    "#989fa3",  # 灰色
]

# 黑夜模式的颜色暂时和light_colors保持一致
dark_colors = [
    "#528d59",  # 绿色
    "#587ad2",  # 蓝色
    "#c24d46",  # 红色
    "#9cbe5d",  # 青绿色
    "#6ebad3",  # 天蓝色
    "#dfb142",  # 橙色
    "#6d4ba4",  # 紫色
    "#8cc5b7",  # 淡青绿色
    "#892d58",  # 紫红色
    "#40877c",  # 深青绿色
    "#d0703c",  # 深橙色
    "#d47694",  # 粉红色
    "#e3b292",  # 淡橙色
    "#b15fbb",  # 浅紫红色
    "#905f4a",  # 棕色
    "#989fa3",  # 灰色
]


def generate_colors(index: Optional[int] = None) -> Tuple[str, str]:
    """
    生成颜色
    :param index: 生成颜色的索引，约定为历史实验数，可以为None
    :return: 生成的颜色，(白天颜色，夜晚颜色)
    """

    if index is None:
        choice_color = random.choice(light_colors)
        return choice_color, choice_color  # 随机返回一个颜色
    else:
        choice_color_light = light_colors[index % len(light_colors)]
        choice_color_dark = dark_colors[index % len(dark_colors)]
        return choice_color_light, choice_color_dark  # 返回对应索引的颜色，如果超出范围则取模



#!/usr/bin/env python
# -*- coding: utf-8 -*-
r"""
@DATE: 2024/5/5 20:22
@File: callback_cloud.py
@IDE: pycharm
@Description:
    云端回调
"""
import json
import sys

from swankit.callback.models import RuntimeInfo, MetricInfo, ColumnInfo
from swankit.core import SwanLabSharedSettings
from swankit.env import create_time
from swankit.log import FONT

from swanlab.api import LoginInfo, create_http, terminal_login
from swanlab.api import get_http
from swanlab.api.upload import upload_logs
from swanlab.api.upload.model import ColumnModel, ScalarModel, MediaModel, FileModel
from swanlab.data.cloud import ThreadPool
from swanlab.data.cloud import UploadType
from swanlab.env import in_jupyter, is_interactive
from swanlab.error import KeyFileError
from swanlab.log import swanlog
from swanlab.package import (
    get_package_version,
    get_package_latest_version,
    get_key,
)
from .callback_local import LocalRunCallback, get_run, SwanLabRunState
from ..api.http import reset_http


def show_button_html(experiment_url):
    """
    用于在jupyter前端显示云端的环境iframe和按钮
    :param experiment_url: 实验链接
    """
    try:
        # noinspection PyPackageRequirements
        from IPython.display import HTML, display

        iframe_h5 = f'<iframe src="{experiment_url}" width=100% height="600" frameborder="no"></iframe>'
        js_code = f"""
        <script>
            function showIframe() {{
                var iframeHtml = '{iframe_h5}';
                document.getElementById('iframeContainer').innerHTML = iframeHtml;
            }}
        </script>
        """

        total_h5 = f"""
<!DOCTYPE html>
<html lang="en">
<head>
    <meta charset="UTF-8">
    <meta name="viewport" content="width=device-width, initial-scale=1.0">
    <title>Show Iframe</title>
    {js_code}
</head>
<body>
    <style>
        .interactive-button {{
            display: flex;
            align-items: center;
            height: 36px;
            border: 0px;
            background-color: #2c8f63;
            color: white;
            padding: 10px 20px;
            transition: background-color 0.3s, transform 0.2s;
        }}

        .interactive-button:hover {{
            background-color: #5cab87;
            cursor: pointer;
        }}

        .interactive-button:active {{ background-color: #217952; transform: scale(0.96); }} </style> <br> <button 
        onclick="showIframe()" class="interactive-button"> <svg style="height: 16px; margin-right: 8px;" viewBox="0 0 
        46 46" fill="none"> <path d="M10.8439 21.1974C10.6414 21.2854 10.4477 21.3925 10.2655 21.5173L10.2069 
        21.5652C10.1839 21.58 10.1625 21.5969 10.1429 21.6159C6.29135 24.6118 4.22831 29.4416 5.32646 34.282C5.94656 
        37.0577 7.50461 39.5348 9.73801 41.2958C11.9714 43.0568 14.7436 43.994 17.5874 43.9495H18.0219C19.8864 
        43.8697 21.7087 43.3694 23.3526 42.486C24.9964 41.6026 26.4193 40.3589 27.5147 38.848C28.61 37.3371 29.3496 
        35.598 29.678 33.761C30.0065 31.9239 29.9153 30.0363 29.4112 28.2395C28.9181 26.4723 27.8919 24.8437 26.9937 
        23.2551C25.4158 20.4653 23.8343 17.6764 22.2492 14.8884C21.7801 14.0647 21.3057 13.2465 20.8419 
        12.4228C20.2315 11.3353 19.2746 10.1519 19.224 8.86183C19.1733 7.57176 20.2235 6.32701 21.5082 
        6.07912C23.9284 5.61801 25.0639 8.24078 25.0693 8.23812C25.363 8.94035 25.9123 9.50489 26.6063 
        9.81764C27.3002 10.1304 28.087 10.168 28.8077 9.92298C29.5283 9.67791 30.1291 9.1684 30.4885 8.49743C30.8479 
        7.82646 30.9392 7.04405 30.7439 6.30835C30.1514 4.37314 28.9133 2.69953 27.2363 1.56656C25.7615 0.511704 
        23.9847 -0.0372109 22.1719 0.00195984C20.9049 0.00893199 19.6532 0.27989 18.4967 0.797557C17.3402 1.31522 
        16.3043 2.06823 15.4551 3.00856C14.49 4.08707 13.7984 5.38193 13.4389 6.78385C13.0794 8.18576 13.0624 9.6536 
        13.3894 11.0635C13.52 11.593 13.6984 12.1095 13.9225 12.6067C14.5595 14.0514 15.4951 15.3681 16.284 
        16.7355C17.2525 18.4147 18.2209 20.0948 19.1893 21.7758C20.1578 23.4568 21.1351 25.1449 22.1213 
        26.8401C22.9209 28.2421 23.7925 29.4682 23.8805 31.1528C23.9175 32.0513 23.7682 32.9479 23.4419 
        33.7859C23.1156 34.6239 22.6194 35.3854 21.9845 36.0223C21.3496 36.6592 20.5897 37.1578 19.7527 
        37.4868C18.9157 37.8157 18.0196 37.9678 17.121 37.9336C14.0024 37.7923 11.6488 35.4814 11.1744 32.4588C10.58 
        28.6419 13.552 26.5469 13.552 26.5469C14.1782 26.1785 14.6497 25.5955 14.8791 24.906C15.1084 24.2166 15.0801 
        23.4673 14.7993 22.7971C14.5186 22.127 14.0044 21.5813 13.3521 21.2611C12.6998 20.941 11.9536 20.8682 11.2517 
        21.0561C11.1174 21.0939 10.9856 21.1402 10.8572 21.1947" fill="white" /> <path d="M42.8101 31.5968C42.8109 
        30.5198 42.7218 29.4445 42.5435 28.3823C42.2663 26.7069 41.7464 25.0808 41.0002 23.5552C40.5524 22.6463 
        39.9874 21.7374 39.1024 21.2417C38.6593 20.9919 38.1589 20.8617 37.6502 20.8639C37.1416 20.8661 36.6423 
        21.0006 36.2013 21.2541C35.7604 21.5077 35.393 21.8716 35.1352 22.3101C34.8775 22.7485 34.7382 23.2466 
        34.7312 23.7552C34.7072 24.8773 35.3149 25.8875 35.768 26.9217C36.5212 28.6453 36.8623 30.5208 36.7642 
        32.3993C36.6661 34.2777 36.1315 36.1075 35.2029 37.7433C35.146 37.8404 35.0952 37.941 35.051 38.0445C34.8623 
        38.4842 34.7635 38.9573 34.7605 39.4358C34.7802 40.1222 35.0356 40.7808 35.4835 41.3011C35.9315 41.8214 
        36.5449 42.1717 37.2207 42.2932C38.8759 42.589 40.1899 41.347 40.8856 39.9609C42.1643 37.3589 42.823 34.4961 
        42.8101 31.5968Z" fill="white" /> <path d="M28.2309 11.8938C28.1761 11.9043 28.1218 11.9176 28.0683 
        11.9338C27.9593 11.9642 27.8611 12.0249 27.7851 12.1088C27.7091 12.1928 27.6584 12.2965 27.6389 
        12.408C27.6193 12.5195 27.6318 12.6343 27.6748 12.7391C27.7178 12.8438 27.7895 12.9343 27.8818 
        12.9999C29.2375 14.0252 30.3809 15.3043 31.2482 16.7662C31.4838 17.1677 31.6888 17.5865 31.8612 
        18.0189C32.0052 18.3921 32.1971 18.8799 32.6822 18.8532C33.0607 18.8346 33.2153 18.512 33.3192 
        18.1895C33.8137 16.5125 33.9678 14.7534 33.7723 13.0159C33.6331 12.0693 33.4155 11.1359 33.122 
        10.2252C33.0775 10.0047 32.9744 9.80029 32.8235 9.6335C32.7273 9.54627 32.6054 9.49262 32.4761 9.4806C32.3468 
        9.46859 32.2171 9.49886 32.1065 9.56687C32.0016 9.65188 31.9115 9.75365 31.8399 9.86806C31.3956 10.4658 
        30.825 10.9581 30.1687 11.3101C29.8377 11.4861 29.4893 11.6272 29.1292 11.7312C28.828 11.8192 28.5215 11.8325 
        28.2309 11.8938Z" fill="white" /> </svg> Display SwanLab Board </button> <br> <div 
        id="iframeContainer"></div> </body> </html>"""

        display(HTML(total_h5))
    except ImportError:
        pass


class CloudRunCallback(LocalRunCallback):
    login_info: LoginInfo = None
    """
    用户登录信息
    """

    def __init__(self, public: bool):
        super(CloudRunCallback, self).__init__()
        self.pool = ThreadPool()
        self.exiting = False
        """
        标记是否正在退出云端环境
        """
        self.public = public

    @classmethod
    def create_login_info(cls):
        """
        发起登录，获取登录信息，执行此方法会覆盖原有的login_info
        """
        key = None
        try:
            key = get_key()
        except KeyFileError:
            pass
        if key is None and not is_interactive():
            raise KeyFileError(
                "api key not configured (no-tty), call `swanlab.login(api_key=[your_api_key])` or set `swanlab.init(mode=\"local\")`."
            )
        return terminal_login(key)

    @staticmethod
    def _get_package_latest_version():
        """
        cloud模式训练开始时，检测package是否为最新版本
        """
        latest_version = get_package_latest_version()
        local_version = get_package_version()
        if latest_version is not None and latest_version != local_version:
            swanlog.info(f"swanlab version {latest_version} is available!  Upgrade: `pip install -U swanlab`")

    def _view_web_print(self):
        self._watch_tip_print()
        http = get_http()
        proj_url, exp_url = http.web_proj_url, http.web_exp_url
        swanlog.info("🏠 View project at " + FONT.blue(FONT.underline(proj_url)))
        swanlog.info("🚀 View run at " + FONT.blue(FONT.underline(exp_url)))
        return exp_url

    def _clean_handler(self):
        run = get_run()
        if run is None:
            return swanlog.debug("SwanLab Runtime has been cleaned manually.")
        if self.exiting:
            return swanlog.debug("SwanLab is exiting, please wait.")
        self._train_finish_print()
        # 如果正在运行
        run.finish() if run.running else swanlog.debug("Duplicate finish, ignore it.")

    def _except_handler(self, tp, val, tb):
        if self.exiting:
            print("")
            swanlog.error("Aborted uploading by user")
            sys.exit(1)
        self._error_print(tp)
        # 结束运行
        get_run().finish(SwanLabRunState.CRASHED, error=self._traceback_error(tb, tp(val)))
        if tp != KeyboardInterrupt:
            print(self._traceback_error(tb, tp(val)), file=sys.stderr)

    def __str__(self):
        return "SwanLabCloudRunCallback"

    def on_init(self, project: str, workspace: str, logdir: str = None, **kwargs) -> int:
        super(CloudRunCallback, self).on_init(project, workspace, logdir)
        if self.login_info is None:
            swanlog.debug("Login info is None, get login info.")
            self.login_info = self.create_login_info()
        # 检测是否有最新的版本
        self._get_package_latest_version()
        http = create_http(self.login_info)
        return http.mount_project(project, workspace, self.public).history_exp_count

    def before_run(self, settings: SwanLabSharedSettings):
        self.settings = settings

    def on_run(self):
        swanlog.install(self.settings.console_dir)
        http = get_http()
        # 注册实验信息
        http.mount_exp(
            exp_name=self.settings.exp_name,
            colors=self.settings.exp_colors,
            description=self.settings.description,
        )

        # 向swanlog注册输出流回调
        def _write_call_call(message):
            self.pool.queue.put((UploadType.LOG, [message]))

        swanlog.set_write_callback(_write_call_call)

        # 注册系统回调
        self._register_sys_callback()
        # 打印信息
        self._train_begin_print()
        swanlog.info("👋 Hi " + FONT.bold(FONT.default(self.login_info.username)) + ", welcome to swanlab!")
        swanlog.info("Syncing run " + FONT.yellow(self.settings.exp_name) + " to the cloud")
        experiment_url = self._view_web_print()
        # 在Jupyter Notebook环境下，显示按钮
        if in_jupyter():
            show_button_html(experiment_url)

    def on_runtime_info_update(self, r: RuntimeInfo):
        # 执行local逻辑，保存文件到本地
        super(CloudRunCallback, self).on_runtime_info_update(r)
        # 添加上传任务到线程池
        rc = r.config.to_dict() if r.config is not None else None
        rm = r.metadata.to_dict() if r.metadata is not None else None
        rr = r.requirements.dumps() if r.requirements is not None else None
        ro = r.conda.dumps() if r.conda is not None else None
        # 不需要json序列化，上传时会自动序列化
        f = FileModel(requirements=rr, config=rc, metadata=rm, conda=ro)
        self.pool.queue.put((UploadType.FILE, [f]))

    def on_column_create(self, column_info: ColumnInfo):
        error = None
        if column_info.error is not None:
            error = {"data_class": column_info.error.got, "excepted": column_info.error.expected}
        # 这里有些比较抽象的地方：
        # 云端版会自动处理不同类型的数据放在不同的组中，所以如果key没有设置成 {section}/{name} 之类的样式，不需要传递section的名称
        # 但是本地版不会，本地版依靠swanlab的处理结果指定列，所以在Data类型上必须定义获取section_name的方法
        # 云端版不需要这样做，因为云端版会自动处理
        # 因为云端版的设计更加先进，云端版对“列”（本地版叫namespace）做了不同的类型标注，但是本地版没有这个概念
        # 所以这里需要判断一下，如果列类型不为SYSTEM且不是 {section}/{name} 之类的格式，就不传递section_name
        if column_info.section_type == "PUBLIC":
            section_name = None if "/" not in column_info.key else column_info.section_name
        elif column_info.section_type == "SYSTEM":
            section_name = column_info.section_name
        else:
            section_name = None
        column = ColumnModel(
            key=column_info.key,
            name=column_info.name,
            cls=column_info.cls,
            config=column_info.config,
            typ=column_info.chart_type.value.column_type,
            section_name=section_name,
            section_type=column_info.section_type,
            error=error,
        )
        self.pool.queue.put((UploadType.COLUMN, [column]))

    def on_metric_create(self, metric_info: MetricInfo):
        super(CloudRunCallback, self).on_metric_create(metric_info)
        # 有错误就不上传
        if metric_info.error:
            return
        metric = metric_info.metric
        key = metric_info.column_info.key
        key_encoded = metric_info.column_info.key_encode
        step = metric_info.metric_step
        epoch = metric_info.metric_epoch
        # 标量折线图
        if metric_info.column_info.chart_type == metric_info.column_info.chart_type.LINE:
            scalar = ScalarModel(metric, key, step, epoch)
            return self.pool.queue.put((UploadType.SCALAR_METRIC, [scalar]))
        # 媒体指标数据

        # -------------------------- 🤡这里是一点小小的💩 --------------------------
        # 要求上传时的文件路径必须带key_encoded前缀
        if metric_info.metric_buffers is not None:
            metric = json.loads(json.dumps(metric))
            for i, d in enumerate(metric["data"]):
                metric["data"][i] = "{}/{}".format(key_encoded, d)
        # ------------------------------------------------------------------------

        media = MediaModel(metric, key, key_encoded, step, epoch, metric_info.metric_buffers)
        self.pool.queue.put((UploadType.MEDIA_METRIC, [media]))

    def on_stop(self, error: str = None):
        # 打印信息
        self._view_web_print()
        run = get_run()
        # 如果正在退出或者run对象为None或者不在云端环境下
        if self.exiting or run is None:
            return swanlog.debug("SwanLab is exiting or run is None, ignore it.")
        state = run.state
        # 标志正在退出（需要在下面的逻辑之前标志）
        self.exiting = True
        sys.excepthook = self._except_handler

        def _():
            # 关闭线程池，等待上传线程完成
            self.pool.finish()
            # 上传错误日志
            if error is not None:
                msg = [{"message": error, "create_time": create_time(), "epoch": swanlog.epoch + 1}]
                upload_logs(msg, level="ERROR")

        FONT.loading("Waiting for uploading complete", _)
        get_http().update_state(state == SwanLabRunState.SUCCESS)
        reset_http()
        # 取消注册系统回调
        self._unregister_sys_callback()
        self.exiting = False



#!/usr/bin/env python
# -*- coding: utf-8 -*-
r"""
@DATE: 2024-03-18 21:28:47
@File: swanlab/data/auth/__init__.py
@IDE: vscode
@Description:
    用户认证模块，当用户登录时，需要验证用户的身份
    这一块因为可能swanlog模块没有初始化，所以需要自己单独打印一下
"""
from .login import terminal_login, code_login



#!/usr/bin/env python
# -*- coding: utf-8 -*-
r"""
@DATE: 2024/6/12 21:11
@File: logout.py
@IDE: pycharm
@Description:
    登出模块
"""
import shutil
import sys

import click
from swankit.log import FONT

from swanlab.env import get_save_dir
from swanlab.package import has_api_key


@click.command()
def logout():
    """Logout to the swanlab cloud."""
    command = FONT.bold("swanlab login")
    if has_api_key():
        # 如果已经是登录状态，那么则询问用户是否确认，如果确认则删除token文件夹
        confirm = input(FONT.swanlab("Are you sure you want to logout? (y/N): "))
        if confirm.lower() == "y":
            try:
                shutil.rmtree(get_save_dir())
                return print(FONT.swanlab("Logout successfully. You can use `" + command + "` to login again."))
            except Exception as e:
                return print(FONT.swanlab("Logout failed. Reason:" + str(e)))
        else:
            return print(FONT.swanlab("Logout canceled."))
    # 如果还未登录，则不做任何处理，并告知用户如何登录
    tip = FONT.swanlab("You are not logged in. If you want to login in, please use `" + command + "` to login.")
    print(tip)
    return sys.exit(1)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
r"""
@DATE: 2024/4/26 16:40
@File: config.py
@IDE: pycharm
@Description:
    SwanLabConfig 配置类
"""
from typing import Any, Union
from collections.abc import MutableMapping
import yaml
import argparse
from swanlab.log import swanlog
import datetime
import math
from typing import Callable, Optional
from swankit.callback import RuntimeInfo
from swanlab.data.modules import Line
import re
import json
from dataclasses import is_dataclass, asdict

BASE_TYPE = (int, float, str, bool)


def json_serializable(obj):
    """
    将传入的字典转换为JSON可序列化格式。
    :raises TypeError: 对象不是JSON可序列化的
    """
    # 如果对象是基本类型，则直接返回
    # 不可以直接使用isinstance，详见issue: https://github.com/SwanHubX/SwanLab/issues/702
    if obj is None:
        return None
    if type(obj) is float and math.isnan(obj):
        return Line.nan
    if type(obj) is float and math.isinf(obj):
        return Line.inf

    for t in BASE_TYPE:
        if type(obj) is t:
            return obj
        # 继承的子类需要转译
        if isinstance(obj, t):
            return t(obj)

    # 将日期和时间转换为字符串
    if isinstance(obj, (datetime.date, datetime.datetime)):
        return obj.isoformat()

    # 对于列表和元组，递归调用此函数
    elif isinstance(obj, (list, tuple)):
        return [json_serializable(item) for item in obj]

    # 对于字典，递归调用此函数处理值，并将key转换为字典
    elif isinstance(obj, dict):
        return {str(key): json_serializable(value) for key, value in obj.items()}

    # 对于可变映射，递归调用此函数处理值，并将key转换为字典
    elif isinstance(obj, MutableMapping):
        return {str(key): json_serializable(value) for key, value in obj.items()}

    try:
        return str(obj)
    except Exception:
        raise TypeError(f"Object: {obj} is not JSON serializable")


def third_party_config_process(data) -> dict:
    """
    对于一些特殊的第三方库的处理，例如omegaconf
    :raises TypeError: 适配的写入的第三方库都没有命中，抛出TypeError
    """
    # 如果是omegaconf的DictConfig，则转换为字典
    try:
        import omegaconf  # noqa

        if isinstance(data, omegaconf.DictConfig):
            return omegaconf.OmegaConf.to_container(data, resolve=True, throw_on_missing=True)
    except ImportError:
        pass

    # 如果是argparse的Namespace，则转换为字典
    if isinstance(data, argparse.Namespace):
        return vars(data)

    # 如果是dataclass类，转换为字典
    if is_dataclass(data):
        return asdict(data)

    raise TypeError


def parse(config) -> dict:
    """
    Check the configuration item and convert it to a JSON serializable format.
    """
    if config is None:
        return {}

    # 1. 第三方配置类型判断与转换
    try:
        return third_party_config_process(config)
    except TypeError:
        pass

    # 2. 将config转换为可被json序列化的字典
    try:
        return json_serializable(config)
    except TypeError:  # noqa
        pass

    # 3. 尝试序列化，序列化成功直接返回
    try:
        return json.loads(json.dumps(config))
    except Exception as e:  # noqa
        # 还失败就没办法了，👋
        raise TypeError(f"config: {config} is not a json serialized dict, error: {e}")


__config_attr__ = ["_SwanLabConfig__config", "_SwanLabConfig__on_setter"]


class SwanLabConfig(MutableMapping):
    """
    The SwanConfig class is used for realize the invocation method of `run.config.lr`.

    Attention:
    The configuration item must be JSON serializable; Cannot set private attributes by `.__xxx`.
    """

    def __init__(
            self,
            config: Union[MutableMapping, argparse.Namespace] = None,
            on_setter: Optional[Callable[[RuntimeInfo], Any]] = None,
    ):
        """
        实例化配置类，如果settings不为None，说明是通过swanlab.init调用的，否则是通过swanlab.config调用的
        """
        # 每一个实例有自己的config
        self.__config = {}
        if config is not None:
            self.__config.update(parse(config))
        self.__on_setter = on_setter

    @staticmethod
    def __fmt_config(config: dict):
        """
        格式化config，值改为value字段，增加desc和sort字段
        """
        # 遍历每一个配置项，值改为value
        sort = 0
        for key, value in config.items():
            config[key] = {"value": value, "desc": "", "sort": sort}
            sort += 1

    def __save(self):
        """
        保存config为dict
        """
        if not self.__on_setter:
            return swanlog.debug("The configuration is not saved because the setter is not set.")
        try:
            # 深度拷贝一次，防止引用传递
            data = yaml.load(yaml.dump(self.__config), Loader=yaml.FullLoader)
        except Exception as e:
            swanlog.error(f"Error occurred when saving config: {e}")
            return
        # 遍历每一个配置项，值改为value，如果是字典，则递归调用
        self.__fmt_config(data)
        r = RuntimeInfo(config=self.__config)
        self.__on_setter(r)
        swanlog.debug(f"Save configuration.")

    # ---------------------------------- 实现对象风格 ----------------------------------

    def __delattr__(self, name: str):
        """
        删除配置项，如果配置项不存在
        """
        # _*__正则开头的属性不允许删除
        if re.match(r"_.*__", name):
            raise AttributeError(f"Attribute '{name}' is private and cannot be deleted")
        try:
            del self.__config[name]
            self.__save()
        except KeyError:
            raise AttributeError(f"You have not deleted '{name}' in the config of the current experiment")

    def __getattr__(self, name: str):
        """
        如果以点号方式访问属性且属性不存在于类中，尝试从配置字典中获取。
        """
        try:
            return self.__config[name]
        except KeyError:
            raise AttributeError(f"You have not get '{name}' in the config of the current experiment")

    def __setattr__(self, name: str, value: Any) -> None:
        """
        Custom setter attribute, user can not set private attributes.
        """
        name = str(name)
        if name in __config_attr__:
            return super().__setattr__(name, value)
        # _*__正则开头的属性不允许设置
        if re.match(r"_.*__", name):
            raise AttributeError(f"Attribute '{name}' is private and cannot be set")
        # 否则应该设置到配置字典中
        self.__config[name] = parse(value)
        self.__save()

    # ---------------------------------- 实现字典风格 ----------------------------------

    def get(self, name: str, default=None):
        """
        Get the value of a configuration item. If the item does not exist, raise AttributeError.
        """
        try:
            return self.__config[name]
        except KeyError:
            return default

    def __delitem__(self, name: str):
        """
        删除配置项，如果配置项不存在,跳过
        """
        try:
            del self.__config[name]
            self.__save()
        except KeyError:
            raise KeyError(f"You have not set '{name}' in the config of the current experiment when deleting")

    def __getitem__(self, name: str):
        """
        以字典方式获取配置项的值
        """
        # 如果self.__dict__中有name属性，则返回
        if not isinstance(name, str):
            raise TypeError(f"Key must be a string, but got {type(name)}")
        # 以_SwanLabConfig__开头，删除
        if name.startswith("_SwanLabConfig__"):
            name = name[15:]
        try:
            return self.__config[name]
        except KeyError:
            raise KeyError(f"You have not get '{name}' in the config of the current experiment")

    def __setitem__(self, name: str, value: Any) -> None:
        """
        Set the value of a configuration item. If the item does not exist, create it.
        User are not allowed to set private attributes.
        """
        name = str(name)
        self.__config[name] = parse(value)
        self.__save()

    def __iter__(self):
        """
        返回配置字典的迭代器。
        """
        return iter(self.__config)

    def __len__(self):
        """
        返回配置项的数量。
        """
        return len(self.__config)

    def __str__(self):
        return str(self.__config)

    # ---------------------------------- 其他函数 ----------------------------------

    def set(self, name: str, value: Any):
        """
        Explicitly set the value of a configuration item and save it.
        Private attributes are not allowed to be set.
        """
        name = str(name)
        self.__config[name] = parse(value)
        self.__save()

    def pop(self, name: str):
        """
        Delete a configuration item; if the item does not exist, skip.
        """
        try:
            t = self.__config[name]
            del self.__config[name]
            self.__save()
            return t
        except KeyError:
            return None

    def update(self, __m: Union[MutableMapping, argparse.Namespace] = None, **kwargs):
        """
        Update the configuration with the key/value pairs from __m, overwriting existing keys.
        """
        if __m is not None:
            for k, v in parse(__m).items():
                self.__config[k] = v
        for k, v in kwargs.items():
            self.__config[k] = parse(v)
        self.__save()

    def clean(self):
        """
        Clean the configuration.
        Attention: This method will reset the instance and instance will not automatically save the configuration.
        """
        self.__config.clear()
        self.__on_setter = None



"""
@author: cunyue
@file: __init__.py.py
@time: 2024/12/3 20:14
@description: 硬件信息采集
"""

from typing import Callable, List, Any, Optional, Tuple

from .cpu import get_cpu_info
from .gpu.nvidia import get_nvidia_gpu_info
from .memory import get_memory_size
from .npu.ascend import get_ascend_npu_info
from .soc.apple import get_apple_chip_info
from .type import HardwareFuncResult, HardwareCollector, HardwareInfo

__all__ = ["get_hardware_info", "HardwareCollector", "HardwareInfo"]


def get_hardware_info() -> Tuple[Optional[Any], List[HardwareCollector]]:
    """
    采集硬件信息，包括CPU、GPU、内存、硬盘等
    """
    monitor_funcs = []
    # 我们希望计算芯片的信息放在最前面，前端展示用
    nvidia = dec_hardware_func(get_nvidia_gpu_info, monitor_funcs)
    ascend = dec_hardware_func(get_ascend_npu_info, monitor_funcs)
    apple = dec_hardware_func(get_apple_chip_info, monitor_funcs)
    c = dec_hardware_func(get_cpu_info, monitor_funcs)
    m = dec_hardware_func(get_memory_size, monitor_funcs)

    info = {
        "memory": m,
        "cpu": c,
        "gpu": {
            "nvidia": nvidia,
        },
        "npu": {
            "ascend": ascend,
        },
        "soc": {
            "apple": apple,
        },
    }
    return filter_none(info, fallback={}), monitor_funcs


def dec_hardware_func(
    func: Callable[[], HardwareFuncResult],
    monitor_funcs: List[HardwareCollector],
) -> Optional[Any]:
    """
    装饰器，用于记录硬件信息采集函数
    """
    x, y = func()
    if y:
        monitor_funcs.append(y)
    return x


def filter_none(data, fallback=None):
    """
    过滤掉字典中值为None的键值对，只对字典有效
    """
    if isinstance(data, dict):
        data = {k: v for k, v in data.items() if v is not None}
        if all(v is None for v in data.values()):
            return fallback
    return data



"""
@author: cunyue
@file: runtime.py
@time: 2024/11/18 15:13
@description: 运行时信息，操作系统信息等
"""

import os
import platform
import socket
import subprocess
import sys


def get_runtime_info():
    return {
        **get_computer_info(),
        **get_python_info(),
        **get_git_info(),
    }


def get_computer_info():
    return {
        "os": platform.platform(),
        "hostname": socket.gethostname(),
        "pid": os.getpid(),
        "cwd": os.getcwd(),
    }


def get_python_info():
    return {
        "python": platform.python_version(),
        "python_verbose": sys.version,
        "executable": sys.executable,
        "command": " ".join(sys.argv),
    }


# ---------------------------------- git信息 ----------------------------------


def get_git_info():
    """获取git信息"""
    return {
        "git_remote": get_remote_url(),
        "git_info": get_git_branch_and_commit(),
    }


def get_remote_url():
    """获取git仓库的远程链接

    Returns
    -------
    str
        git remote url
    """
    try:
        # 运行git命令获取远程仓库URL
        result = subprocess.run(
            ["git", "config", "--get", "remote.origin.url"], stdout=subprocess.PIPE, stderr=subprocess.PIPE, text=True
        )

        # 检查命令是否成功运行
        if result.returncode == 0:
            url = result.stdout.strip().replace("git@", "https://")
            if url.endswith(".git"):
                url = url[:-4]
            return replace_second_colon(url, "/")
        else:
            return None
    except Exception as e:  # noqa
        return None


def replace_second_colon(input_string, replacement):
    """替换字符串中第二个‘:’"""
    first_colon_index = input_string.find(":")
    if first_colon_index != -1:
        second_colon_index = input_string.find(":", first_colon_index + 1)
        if second_colon_index != -1:
            return input_string[:second_colon_index] + replacement + input_string[second_colon_index + 1 :]
    return input_string


def get_git_branch_and_commit():
    """获取项目git的分支名和该分支下最新提交的hash"""
    try:
        # 获取当前分支名称
        branch_process = subprocess.Popen(
            ["git", "branch", "--show-current"], stdout=subprocess.PIPE, stderr=subprocess.PIPE, text=True
        )
        branch_name, branch_err = branch_process.communicate()
        if branch_process.returncode != 0:
            return None, None

        branch_name = branch_name.strip()

        # 获取当前分支的最新提交hash
        commit_process = subprocess.Popen(
            ["git", "rev-parse", branch_name], stdout=subprocess.PIPE, stderr=subprocess.PIPE, text=True
        )
        commit_hash, commit_err = commit_process.communicate()

        # 如果无法获取最新提交hash, 那么就返回branch_name和None
        if commit_process.returncode != 0:
            return branch_name, None

        commit_hash = commit_hash.strip()
        return branch_name, commit_hash

    except Exception as e:  # noqa
        return None, None



#!/usr/bin/env python
# -*- coding: utf-8 -*-
r"""
@DATE: 2024/5/5 20:23
@File: callback_local.py
@IDE: pycharm
@Description:
    基本回调函数注册表，此时不考虑云端情况
"""
import json
import os
import sys
import traceback
from datetime import datetime
from typing import Union

from swankit.callback.models import RuntimeInfo, MetricInfo
from swankit.core import SwanLabSharedSettings
from swankit.log import FONT

from swanlab.data.run.callback import SwanLabRunCallback
from swanlab.data.run.main import get_run, SwanLabRunState
from swanlab.env import SwanLabEnv
from swanlab.log import swanlog


class LocalRunCallback(SwanLabRunCallback):

    def __init__(self):
        super(LocalRunCallback, self).__init__()

    @staticmethod
    def _traceback_error(tb, val):
        """
        获取traceback信息
        """
        trace_list = traceback.format_tb(tb)
        html = ""
        for line in trace_list:
            html += line
        html += str(val)
        return html

    @staticmethod
    def _error_print(tp):
        """
        错误打印
        """
        # 如果是KeyboardInterrupt异常
        if tp == KeyboardInterrupt:
            swanlog.info("KeyboardInterrupt by user")
        else:
            swanlog.info("Error happened while training")

    @staticmethod
    def _init_logdir(logdir: Union[str, None] = None) -> None:
        """
        根据传入的logdir,初始化日志文件夹
        ---
        Args:
            logdir: 日志文件夹路径

        Return:
            None

        Step:
            1: 参数检查
            2: 环境变量设置
            3: 默认路径
            4: .gitignore
        """
        env_key = SwanLabEnv.SWANLOG_FOLDER.value
        # 如果传入了logdir，则将logdir设置为环境变量，代表日志文件存放的路径
        # 如果没有传入logdir，则使用默认的logdir, 即当前工作目录下的swanlog文件夹，但是需要保证目录存在
        if logdir is None:
            logdir = os.environ.get(env_key) or os.path.join(os.getcwd(), "swanlog")

        logdir = os.path.abspath(logdir)
        try:
            os.makedirs(logdir, exist_ok=True)
            if not os.access(logdir, os.W_OK):
                raise IOError(f"no write permission for path: {logdir}")
        except Exception as error:
            raise IOError(f"Failed to create or access logdir: {logdir}, error: {error}")

        os.environ[env_key] = logdir

        # 如果logdir是空的，创建.gitignore文件，写入*
        if not os.listdir(logdir):
            with open(os.path.join(logdir, ".gitignore"), "w", encoding="utf-8") as f:
                f.write("*")

    def __str__(self):
        return "SwanLabLocalRunCallback"

    def _except_handler(self, tp, val, tb):
        """
        异常处理
        """
        self._error_print(tp)
        # 结束运行
        get_run().finish(SwanLabRunState.CRASHED, error=self._traceback_error(tb, tp(val)))
        if tp != KeyboardInterrupt:
            print(self._traceback_error(tb, tp(val)), file=sys.stderr)

    def _clean_handler(self):
        run = get_run()
        if run is None:
            return swanlog.debug("SwanLab Runtime has been cleaned manually.")
        self._train_finish_print()
        # 如果正在运行
        run.finish() if run.running else swanlog.debug("Duplicate finish, ignore it.")

    def on_init(self, proj_name: str, workspace: str, logdir: str = None, **kwargs):
        self._init_logdir(logdir)

    def before_run(self, settings: SwanLabSharedSettings):
        self.settings = settings

    def on_run(self):
        swanlog.install(self.settings.console_dir)
        # 注入系统回调
        self._register_sys_callback()
        # 打印信息
        self._train_begin_print()
        swanlog.info("Experiment_name: " + FONT.yellow(self.settings.exp_name))
        self._watch_tip_print()

    def on_runtime_info_update(self, r: RuntimeInfo):
        # 更新运行时信息
        if r.requirements is not None:
            r.requirements.write(self.settings.files_dir)
        if r.metadata is not None:
            r.metadata.write(self.settings.files_dir)
        if r.config is not None:
            r.config.write(self.settings.files_dir)

    def on_metric_create(self, metric_info: MetricInfo):
        # 出现任何错误直接返回
        if metric_info.error:
            return
        # 屏蔽非自定义指标，因为现在本地不支持系统指标
        if metric_info.column_info.cls != "CUSTOM":
            return
        # ---------------------------------- 保存指标数据 ----------------------------------
        self.settings.mkdir(os.path.dirname(metric_info.metric_file_path))
        self.settings.mkdir(os.path.dirname(metric_info.summary_file_path))
        with open(metric_info.summary_file_path, "w+", encoding="utf-8") as f:
            f.write(json.dumps(metric_info.metric_summary, ensure_ascii=False))
        with open(metric_info.metric_file_path, "a", encoding="utf-8") as f:
            f.write(json.dumps(metric_info.metric, ensure_ascii=False) + "\n")

        # ---------------------------------- 保存媒体字节流数据 ----------------------------------
        if metric_info.metric_buffers is None:
            return
        for i, r in enumerate(metric_info.metric_buffers):
            if r is None:
                continue
            # 组合路径
            path = os.path.join(metric_info.swanlab_media_dir, metric_info.column_info.kid)
            os.makedirs(path, exist_ok=True)
            # 写入数据
            with open(os.path.join(path, metric_info.metric["data"][i]), "wb") as f:
                f.write(r.getvalue())

    def on_stop(self, error: str = None):
        """
        训练结束，取消系统回调
        此函数被`run.finish`调用
        """
        # 写入错误信息
        if error is not None:
            with open(self.settings.error_path, "a") as fError:
                print(datetime.now(), file=fError)
                print(error, file=fError)
        # 打印信息
        self._watch_tip_print()
        # 取消注册系统回调
        self._unregister_sys_callback()



#!/usr/bin/env python
# -*- coding: utf-8 -*-
r"""
@DATE: 2024-03-20 14:41:25
@File: swanlab/auth/info.py
@IDE: vscode
@Description:
    定义认证数据格式
"""
from typing import Union

import requests

from swanlab.package import save_key, fmt_web_host


class LoginInfo:
    """
    登录信息类，负责解析登录接口返回的信息，并且进行保存
    无论接口请求成功还是失败，都会初始化一个LoginInfo对象
    """

    def __init__(self, resp: requests.Response, api_key: str, api_host: str, web_host: str):
        self.__api_key = api_key
        self.__resp = resp
        self.__body = resp.json() if resp.status_code == 200 else {}
        self.__username = None
        """
        如果此属性不为None，username返回此属性
        """
        self.web_host = fmt_web_host(web_host)
        self.api_host = api_host

    @property
    def sid(self) -> Union[str, None]:
        """
        获取sid，如果请求失败则返回None
        """
        return self.__body.get("sid")

    @property
    def expired_at(self) -> Union[str, None]:
        """
        获取过期时间，如果请求失败则返回None
        """
        return self.__body.get("expiredAt")

    @property
    def username(self) -> Union[str, None]:
        """
        获取用户名，如果请求失败则返回None
        """
        if self.__username is not None:
            return self.__username
        return self.__body.get("userInfo", {}).get("username")

    @username.setter
    def username(self, value):
        """
        设置用户名
        """
        self.__username = value

    @property
    def is_fail(self):
        """
        判断登录是否失败
        """
        return self.__resp.status_code != 200

    @property
    def api_key(self):
        """
        获取api_key
        """
        if self.is_fail:
            return None
        return self.__api_key

    def __str__(self) -> str:
        """错误时会返回错误信息"""
        if self.__resp.reason == "OK":
            return "Login success"
        if self.__resp.reason == "Unauthorized" or self.__resp.reason == "Authorization Required":
            return "Error api key"
        if self.__resp.reason == "Forbidden":
            return "You need to be verified first"
        return str(self.__resp.status_code) + " " + self.__resp.reason

    def save(self):
        """
        保存登录信息
        """
        return save_key(self.web_host, self.api_key, self.api_host)


class ProjectInfo:
    def __init__(self, data: dict):
        self.__data = data

    @property
    def cuid(self):
        return self.__data["cuid"]

    @property
    def name(self):
        return self.__data["name"]

    @property
    def history_exp_count(self):
        return self.__data.get('_count', {'experiments': 0})["experiments"]


class ExperimentInfo:

    def __init__(self, data: dict):
        self.__data = data

    @property
    def cuid(self):
        return self.__data["cuid"]

    @property
    def name(self):
        return self.__data["name"]



import math
import yaml
from swanlab.data.run.main import SwanLabRun, get_run, swanlog, get_config
from swanlab.data.run.config import SwanLabConfig, parse, Line, RuntimeInfo, MutableMapping
import pytest
import omegaconf
from dataclasses import dataclass
import argparse


@pytest.fixture(scope="function", autouse=True)
def setup_function():
    """
    在当前测试文件下的每个测试函数执行前后执行
    """
    if get_run() is not None:
        get_run().finish()
    swanlog.disable_log()
    yield
    swanlog.enable_log()


def test_parse():
    """
    测试config.parse函数
    """
    # ---------------------------------- omegaConf ----------------------------------
    config_data = {
        "a": 1,
        "b": "mnist",
        "c/d": [1, 2, 3],
        "e/f/h": {"a": 1, "b": {"c": 2}},
    }
    cfg = omegaconf.OmegaConf.create(config_data)
    config = parse(cfg)
    assert yaml.dump(config) == yaml.dump(config_data)

    # ---------------------------------- 自定义继承自MutableMapping的类 ----------------------------------

    class Test(MutableMapping):
        def __init__(self, a, b):
            self.data = {"a": a, "b": b}

        def __setitem__(self, __key, __value):
            self.data[__key] = __value

        def __delitem__(self, __key):
            del self.data[__key]

        def __getitem__(self, __key):
            return self.data.get(__key, None)

        def __len__(self):
            return len(self.data)

        def __iter__(self):
            return iter(self.data)

    config_data = {"a": 1, "b": "mnist", "c/d": [1, 2, 3], "e/f/h": {"a": 1, "b": {"c": 2}}, "test": Test(1, 2)}
    config = parse(config_data)
    assert config["test"]["a"] == 1
    assert config["test"]["b"] == 2
    # ---------------------------------- 包含NaN或者INF的dict对象 ----------------------------------
    config_data = {
        "inf": math.inf,
        "nan": math.nan,
    }
    config = parse(config_data)
    assert config["inf"] == Line.inf
    assert config["nan"] == Line.nan

    # ---------------------------------- dataclass support ----------------------------------
    @dataclass
    class MyData:
        a: int
        b: float

    config_data = MyData(10, 20.0)
    config = parse(config_data)
    assert config["a"] == 10
    assert config["b"] == 20.0
    # ---------------------------------- argparse.Namespace ----------------------------------
    config_data = argparse.Namespace(a=1, b="mnist", c=[1, 2, 3], d={"a": 1, "b": {"c": 2}})
    config = parse(config_data)
    assert yaml.dump(config) == yaml.dump(vars(config_data))


def test_parse_base_class():
    """
    继承自基础的类不能绕过parse函数
    """

    class StrChild(str):
        pass

    value = parse(StrChild("abc"))
    assert value == "abc"
    assert yaml.safe_dump({"value": value})

    class IntChild(int):
        pass

    value = parse(IntChild(1))
    assert value == 1
    assert value != "1"
    assert yaml.safe_dump({"value": value})


class TestSwanLabConfigOperation:
    """
    单独测试TestSwanLabRunConfig这个类
    """

    def test_basic_operation_object(self):
        """
        测试类的基本操作，增删改
        """
        config = SwanLabConfig()
        # ---------------------------------- 对象风格设置 ----------------------------------
        config.a = 1
        assert config.a == 1
        config.a = 2
        assert config.a == 2
        with pytest.raises(AttributeError):
            config.__a = 1
        # 不存在的属性
        with pytest.raises(AttributeError):
            config.b  # noqa
        # 删除属性
        del config.a
        with pytest.raises(AttributeError):
            del config.a  # 重复删除报错
        with pytest.raises(AttributeError):
            config.a  # noqa

    def test_basic_operation_dict(self):
        """
        测试字典的基本操作，增删改
        """
        config = SwanLabConfig()
        # ---------------------------------- 字典风格设置 ----------------------------------
        config["a"] = 1
        assert config["a"] == 1
        config["a"] = 2
        assert config["a"] == 2
        # 删除属性
        del config["a"]
        with pytest.raises(KeyError):
            del config["a"]
        with pytest.raises(KeyError):
            config["a"]  # noqa
        # 字典风格可以设置，读取，删除私有属性
        config["__a"] = 1
        assert config["__a"] == 1
        del config["__a"]
        with pytest.raises(KeyError):
            del config["__a"]  # 重复删除失败
        with pytest.raises(KeyError):
            config["__a"]  # noqa
        # int访问，设置
        config[1] = 1  # noqa
        with pytest.raises(TypeError):
            assert config[1] == 1  # noqa

    def test_dict_iter(self):
        """
        测试字典风格的迭代
        """
        config = SwanLabConfig()
        ll = ["a", "b", "c", "d"]
        for i in ll:
            config[i] = i
        assert set(config) == {"a", "b", "c", "d"}
        index = 0
        # 返回顺序相同
        for key in config:
            assert key == ll[index]
            index += 1

    def test_dict_len(self):
        """
        测试字典风格的长度
        """
        config = SwanLabConfig()
        assert len(config) == 0
        config["a"] = 1
        assert len(config) == 1
        config["b"] = 2
        assert len(config) == 2
        del config["a"]
        assert len(config) == 1
        del config["b"]
        assert len(config) == 0

    def test_func_operation(self):
        """
        测试内置函数操作
        """
        config = SwanLabConfig()
        # ---------------------------------- get ----------------------------------
        a = config.get("a")
        assert a is None
        a = config.get("a", 1)
        assert a == 1
        config["a"] = 5
        a = config.get("a")
        assert a == 5
        # ---------------------------------- set ----------------------------------
        config.set("b", 1)
        assert config["b"] == 1
        config.set("__b", 1)
        assert config["__b"] == 1
        # ---------------------------------- pop ----------------------------------
        config["c"] = 9
        c = config.pop("c")
        assert c == 9
        assert config.pop("d") is None
        # ---------------------------------- clean ----------------------------------
        config["e"] = 1
        config["g"] = 0
        config.clean()
        assert len(config) == 0
        with pytest.raises(KeyError):
            config["e"]  # noqa
        # ---------------------------------- update ----------------------------------
        config["x"] = 1
        config["y"] = 2
        config["z"] = {"a": 1, "b": 2}
        config.update({"x": 2, "y": 3, "z": {"a": 2, "b": 3}})
        assert config["x"] == 2
        assert config["y"] == 3
        assert config["z"] == {"a": 2, "b": 3}
        # update自己
        _config = SwanLabConfig()
        _config.update(config)
        assert _config == config
        # update，argparse.Namespace
        _config = SwanLabConfig()
        _config.update(argparse.Namespace(a=1, b=2))
        assert _config["a"] == 1
        assert _config["b"] == 2
        # update, use kwargs
        _config = SwanLabConfig()
        _config.update(a=2, b=1)
        assert _config["a"] == 2
        assert _config["b"] == 1


def test_on_setter():
    """
    测试on_setter函数，在设置属性时触发
    """
    num = 1

    def on_setter(_: RuntimeInfo):
        nonlocal num
        num += 1

    config = SwanLabConfig(on_setter=on_setter)

    # ---------------------------------- 对象、字典风格 ----------------------------------

    # 设置触发
    config.a = 1
    assert num == 2
    del config.a
    assert num == 3
    config["b"] = 1
    assert num == 4
    del config["b"]
    assert num == 5

    # 读取不触发
    config.x = 1
    assert num == 6
    _ = config.x
    assert num == 6
    config["y"] = 1
    assert num == 7
    _ = config["y"]
    assert num == 7

    # ---------------------------------- api ----------------------------------

    # 设置触发
    config.set("c", 1)
    assert num == 8
    config.pop("c")
    assert num == 9
    config.update({"d": {}})
    assert num == 10
    # 深层设置无法触发
    config.d["e"] = 1
    assert num == 10

    # 读取不触发
    config.get("f", 1)
    config["g"] = 1
    assert num == 11
    config.get("g")
    assert num == 11

    # ---------------------------------- clean以后再设置无法触发 ----------------------------------

    config.clean()
    config.h = 1
    assert num == 11


class TestSwanLabConfigWithRun:
    """
    测试SwanLabConfig与SwanLabRun的交互
    """

    def test_use_dict(self):
        """
        正常流程，输入字典
        """
        run = SwanLabRun(
            run_config={
                "a": 1,
                "b": "mnist",
                "c/d": [1, 2, 3],
                "e/f/h": {"a": 1, "b": {"c": 2}},
            }
        )
        config = run.config
        _config = get_config()
        assert config["a"] == _config["a"] == 1
        assert config["b"] == _config["b"] == "mnist"
        assert config["c/d"] == _config["c/d"] == [1, 2, 3]

    def test_use_omegaconf(self):
        """
        正常流程，输入OmegaConf
        """
        run = SwanLabRun(
            run_config=omegaconf.OmegaConf.create(
                {
                    "a": 1,
                    "b": "mnist",
                    "c/d": [1, 2, 3],
                    "e/f/h": {"a": 1, "b": {"c": 2}},
                }
            )
        )
        config = run.config
        _config = get_config()
        assert config["a"] == _config["a"] == 1
        assert config["b"] == _config["b"] == "mnist"
        assert config["c/d"] == _config["c/d"] == [1, 2, 3]

    def test_use_argparse(self):
        """
        正常流程，输入argparse.Namespace
        """
        run = SwanLabRun(run_config=argparse.Namespace(a=1, b="mnist", c=[1, 2, 3], d={"a": 1, "b": {"c": 2}}))
        config = run.config
        _config = get_config()
        assert config["a"] == _config["a"] == 1
        assert config["b"] == _config["b"] == "mnist"
        assert config["c"] == _config["c"] == [1, 2, 3]

    def test_use_config(self):
        """
        正常流程，输入SwanLabConfig
        """
        run = SwanLabRun(
            run_config=SwanLabConfig(
                {
                    "a": 1,
                    "b": "mnist",
                    "c": [1, 2, 3],
                    "e/f/h": {"a": 1, "b": {"c": 2}},
                }
            )
        )
        config = run.config
        _config = get_config()
        assert config["a"] == _config["a"] == 1
        assert config["b"] == _config["b"] == "mnist"
        assert config["c"] == _config["c"] == [1, 2, 3]

    def test_after_finish(self):
        """
        测试在finish之后config的变化
        """
        run = SwanLabRun(
            run_config={
                "a": 1,
                "b": "mnist",
                "c/d": [1, 2, 3],
                "e/f/h": {"a": 1, "b": {"c": 2}},
            }
        )
        run.finish()
        config = run.config
        _config = get_config()
        assert len(config) == 4
        assert len(_config) == 0

    def test_error_config_input(self):
        """
        测试错误的输入
        """
        with pytest.raises(TypeError):
            SwanLabRun(run_config=1)
        with pytest.raises(TypeError):
            SwanLabRun(run_config="1")
        with pytest.raises(TypeError):
            SwanLabRun(run_config=[1, 2, 3])
        with pytest.raises(TypeError):
            SwanLabRun(run_config=(1, 2, 3))
        with pytest.raises(TypeError):
            SwanLabRun(run_config=True)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
r"""
@DATE: 2024/4/7 16:51
@File: http.py
@IDE: pycharm
@Description:
    http会话对象
"""
import json
from datetime import datetime
from typing import Optional, Tuple, Dict, Union, List, AnyStr

import requests
from requests.adapters import HTTPAdapter
from swankit.log import FONT
from urllib3.util.retry import Retry

from swanlab.data.modules import MediaBuffer
from swanlab.error import NetworkError, ApiError
from swanlab.log import swanlog
from swanlab.package import get_package_version
from .auth.login import login_by_key
from .cos import CosClient
from .info import LoginInfo, ProjectInfo, ExperimentInfo


def decode_response(resp: requests.Response) -> Union[Dict, AnyStr]:
    """
    解码响应，返回信息
    低版本requests库没有JSONDecodeError，所以需要捕获两种异常
    """
    try:
        return resp.json()
    except json.decoder.JSONDecodeError:
        return resp.text
    except requests.JSONDecodeError:
        return resp.text


class HTTP:
    """
    封装请求函数，添加get、post、put、delete方法
    """

    REFRESH_TIME = 60 * 60 * 24 * 7  # 7天
    """
    刷新时间，单位秒，如果sid过期时间减去当前时间小于这个时间，就刷新sid
    """

    def __init__(self, login_info: LoginInfo):
        """
        初始化会话
        """
        self.__login_info = login_info
        # 当前cos信息
        self.__cos: Optional[CosClient] = None
        # 当前项目信息
        self.__proj: Optional[ProjectInfo] = None
        # 当前实验信息
        self.__exp: Optional[ExperimentInfo] = None
        # 当前进程会话
        self.__session: Optional[requests.Session] = None
        # 当前项目所属的username
        self.__username = login_info.username
        self.__version = get_package_version()
        # 创建会话
        self.__create_session()

    # ---------------------------------- 一些辅助属性 ----------------------------------

    @property
    def base_url(self):
        return self.__login_info.api_host

    @property
    def api_host(self):
        return self.__login_info.web_host

    @property
    def web_host(self):
        return self.__login_info.web_host

    @property
    def groupname(self):
        """
        当前项目所属组名
        """
        return self.__username

    @property
    def username(self):
        """
        当前登录的用户名
        """
        return self.__login_info.username

    @property
    def cos(self):
        return self.__cos

    @property
    def proj_id(self):
        return self.__proj.cuid

    @property
    def projname(self):
        return self.__proj.name

    @property
    def exp_id(self):
        return self.__exp.cuid

    @property
    def expname(self):
        return self.__exp.name

    @property
    def sid_expired_at(self):
        """
        获取sid的过期时间，字符串格式转时间
        """
        return datetime.strptime(self.__login_info.expired_at, "%Y-%m-%dT%H:%M:%S.%fZ")

    @property
    def web_proj_url(self):
        return f"{self.web_host}/@{self.groupname}/{self.projname}"

    @property
    def web_exp_url(self):
        return f"{self.web_proj_url}/runs/{self.exp_id}"

    # ---------------------------------- http方法 ----------------------------------

    def __before_request(self):
        """
        请求前的钩子
        """
        # FIXME datetime.utcnow() -> datetime.now(datetime.UTC)
        if (self.sid_expired_at - datetime.utcnow()).total_seconds() <= self.REFRESH_TIME:
            # 刷新sid，新建一个会话
            swanlog.debug("Refresh sid...")
            self.__login_info = login_by_key(self.__login_info.api_key, save=False)
            self.__session.headers["cookie"] = f"sid={self.__login_info.sid}"

    def __create_session(self):
        """
        创建会话，这将在HTTP类实例化时调用
        添加了重试策略
        """
        session = requests.Session()
        retry = Retry(
            total=3,
            backoff_factor=0.1,
            status_forcelist=[429, 500, 502, 503, 504],
            allowed_methods=frozenset(["GET", "POST", "PUT", "DELETE", "PATCH"]),
        )
        adapter = HTTPAdapter(max_retries=retry)
        session.mount("https://", adapter)

        session.headers["swanlab-sdk"] = self.__version
        session.cookies.update({"sid": self.__login_info.sid})

        # 注册响应钩子
        def response_interceptor(response: requests.Response, *args, **kwargs):
            """
            捕获所有的http不为2xx的错误，以ApiError的形式抛出
            """
            if response.status_code // 100 != 2:
                raise ApiError(response, response.status_code, response.reason)

        session.hooks["response"] = response_interceptor

        self.__session = session

    def post(self, url: str, data: dict = None) -> Union[dict, str]:
        """
        post请求
        """
        url = self.base_url + url
        self.__before_request()
        resp = self.__session.post(url, json=data)
        return decode_response(resp)

    def put(self, url: str, data: dict = None) -> Union[dict, str]:
        """
        put请求
        """
        url = self.base_url + url
        self.__before_request()
        resp = self.__session.put(url, json=data)
        return decode_response(resp)

    def get(self, url: str, params: dict = None) -> dict:
        """
        get请求
        """
        url = self.base_url + url
        self.__before_request()
        resp = self.__session.get(url, params=params)
        return decode_response(resp)

    def patch(self, url: str, data: dict = None) -> Union[dict, str]:
        """
        patch请求
        """
        url = self.base_url + url
        self.__before_request()
        resp = self.__session.patch(url, json=data)
        return decode_response(resp)

    # ---------------------------------- 对象存储方法 ----------------------------------

    def __get_cos(self):
        cos = self.get(f"/project/{self.groupname}/{self.projname}/runs/{self.exp_id}/sts")
        self.__cos = CosClient(cos)

    def upload(self, buffer: MediaBuffer):
        """
        上传文件，需要注意的是file_path应该为unix风格而不是windows风格
        :param buffer: 自定义文件内存对象
        """
        if self.__cos.should_refresh:
            self.__get_cos()
        return self.__cos.upload(buffer)

    def upload_files(self, buffers: List[MediaBuffer]) -> Dict[str, Union[bool, List]]:
        """
        批量上传文件，keys和local_paths的长度应该相等
        :param buffers: 文件内存对象
        :return: 返回上传结果, 包含success_all和detail两个字段，detail为每一个文件的上传结果（通过index索引对应）
        """
        if self.__cos.should_refresh:
            swanlog.debug("Refresh cos...")
            self.__get_cos()
        return self.__cos.upload_files(buffers)

    # ---------------------------------- 接入后端api ----------------------------------

    def mount_project(self, name: str, username: str = None, public: bool = None) -> ProjectInfo:
        """
        创建项目，如果项目已存在，则获取项目信息
        :param name: 项目名称
        :param username: 项目所属的用户名
        :param public: 项目是否公开
        :return: 项目信息
        """
        self.__username = self.__username if username is None else username

        def _():
            try:
                visibility = "PUBLIC" if public else "PRIVATE"
                resp = http.post(f"/project/{self.groupname}", data={"name": name, "visibility": visibility})
            except ApiError as e:
                # 如果为409，表示已经存在，获取项目信息
                if e.resp.status_code == 409:
                    resp = http.get(f"/project/{http.groupname}/{name}")
                elif e.resp.status_code == 404:
                    # 组织/用户不存在
                    raise ValueError(f"Entity `{http.groupname}` not found")
                elif e.resp.status_code == 403:
                    # 权限不足
                    raise ValueError(f"Entity permission denied: " + http.groupname)
                else:
                    raise e
            return ProjectInfo(resp)

        project: ProjectInfo = FONT.loading("Getting project...", _)
        self.__proj = project
        return project

    def mount_exp(self, exp_name, colors: Tuple[str, str], description: str = None):
        """
        初始化实验，获取存储信息
        :param exp_name: 所属实验名称
        :param colors: 实验颜色，有两个颜色
        :param description: 实验描述
        """

        def _():
            """
            先创建实验，后生成cos凭证
            :return:
            """

            data = self.post(
                f"/project/{self.groupname}/{self.__proj.name}/runs",
                (
                    {"name": exp_name, "colors": list(colors), "description": description}
                    if description
                    else {"name": exp_name, "colors": list(colors)}
                ),
            )
            self.__exp = ExperimentInfo(data)
            # 获取cos信息
            self.__get_cos()

        FONT.loading("Creating experiment...", _)

    def update_state(self, success: bool):
        """
        更新实验状态
        :param success: 实验是否成功
        """

        def _():
            self.put(
                f"/project/{self.groupname}/{self.projname}/runs/{self.exp_id}/state",
                {"state": "FINISHED" if success else "CRASHED", "from": "sdk"},
            )

        FONT.loading("Updating experiment status...", _)


http: Optional["HTTP"] = None
"""
一个进程只有一个http请求对象
"""


def create_http(login_info: LoginInfo) -> HTTP:
    """
    创建http请求对象
    """
    global http
    http = HTTP(login_info)
    return http


def get_http() -> HTTP:
    """
    创建http请求对象
    :return: http请求对象
    """
    global http
    if http is None:
        raise ValueError("http object is not initialized")
    return http


def reset_http():
    """
    重置http对象
    """
    global http
    http = None


def sync_error_handler(func):
    """
    在一些接口中我们不希望线程奔溃，而是返回一个错误对象
    """

    def wrapper(*args, **kwargs) -> Tuple[Optional[Union[dict, str]], Optional[Exception]]:
        try:
            # 在装饰器中调用被装饰的异步函数
            result = func(*args, **kwargs)
            return result, None
        except requests.exceptions.Timeout:
            return None, NetworkError()
        except requests.exceptions.ConnectionError:
            return None, NetworkError()
        except Exception as e:
            return None, e

    return wrapper



"""
Docs: <WAITIGNG UPDATE>

For adaptation to the huggingface accelerate. You can used SwanLab as your tracker, experiment logs can be uploaded to
SwanLab or viewed using the local version of SwanLab. Detailed of used swanlab in accelerate train scripts are as follows:
------train.py in accelerate------
...
from swanlab.integration.accelerate import SwanLabTracker
...
tracker = SwanLabTracker("some_run_name")
accelerator = Accelerator(log_with=tracker)
...
---------------------------------
These also can be mixed with existing trackers, including with "all":
------train.py in accelerate------
...
from swanlab.integration.accelerate import SwanLabTracker
...
tracker = SwanLabTracker("some_run_name")
accelerator = Accelerator(log_with=[tracker, "all"])
...
---------------------------------
"""

import os
from typing import Any, Dict, List, Optional, Union

import swanlab

try:
    from accelerate.tracking import GeneralTracker, on_main_process
    from accelerate.logging import get_logger
except ImportError:
    raise RuntimeError(
        "This contrib module requires Accelerate to be installed. "
        "Please install it with command: \n pip install accelerate"
    )


logger = get_logger(__name__)


class SwanLabTracker(GeneralTracker):
    """
    A `Tracker` class that supports `swanlab`. Should be initialized at the start of your script.

    Args:
        run_name (`str`):
            The name of the experiment run
        logging_dir (`str`, `os.PathLike`):
            Location for swanlab logs to be stored.
        kwargs:
            Additional key word arguments passed along to the `swanlab.init` method.
    """

    name = "swanlab"
    requires_logging_directory = False

    @on_main_process
    def __init__(
        self,
        project: Optional[str] = None,
        workspace: Optional[str] = None,
        experiment_name: Optional[str] = None,
        description: Optional[str] = None,
        logdir: Optional[str] = None,
        mode: Optional[str] = None,
        **kwargs,
    ):
        super().__init__()

        self._swanlab_init: Dict[str, Any] = {
            "project": project,
            "workspace": workspace,
            "experiment_name": experiment_name,
            "description": description,
            "logdir": logdir,
            "mode": mode,
        }

        self._swanlab_init.update(**kwargs)

        self._project = self._swanlab_init.get("project")
        self._workspace = self._swanlab_init.get("workspace")
        self._experiment_name = self._swanlab_init.get("experiment_name")
        self._description = self._swanlab_init.get("decsription")
        self._logdir = self._swanlab_init.get("logdir")
        self._mode = self._swanlab_init.get("mode")

        self.logdir = os.path.join(logdir, self._project) if self._logdir is not None else None

        swanlab.config["FRAMEWORK"] = "accelerate"
        if swanlab.get_run() is None:
            self.writer = swanlab.init(**self._swanlab_init)
        else:
            self.writer = swanlab.get_run()

        logger.debug(f"Initialized swanlab project {self._project} logging to {self._logdir}")
        logger.debug(
            "Make sure to log any initial configurations with `self.store_init_configuration` before training!"
        )

    def update_config(self, config: Dict[str, Any]):
        swanlab.config.update(config)

    @property
    def tracker(self):
        return self.writer

    @on_main_process
    def store_init_configuration(self, values: dict):
        """
        Logs `values` as hyperparameters for the run. Should be run at the beginning of your experiment. Stores the
        hyperparameters in a yaml file for future use.

        Args:
            values (Dictionary `str` to `bool`, `str`, `float` or `int`):
                Values to be stored as initial hyperparameters as key-value pairs. The values need to have type `bool`,
                `str`, `float`, `int`, or `None`.
        """
        swanlab.config.update(values)
        logger.debug("Stored initial configuration hyperparameters to swanlab")

    @on_main_process
    def log(self, values: dict, step: Optional[int] = None, **kwargs):
        """
        Logs `values` to the current run.

        Args:
            values (Dictionary `str` to `str`, `float`, `int` or `dict` of `str` to `float`/`int`):
                Values to be logged as key-value pairs. The values need to have type `str`, `float`, `int` or `dict` of
                `str` to `float`/`int`.
            step (`int`, *optional*):
                The run step. If included, the log will be affiliated with this step.
            kwargs:
                Additional key word arguments passed along to either `SummaryWriter.add_scaler`,
                `SummaryWriter.add_text`, or `SummaryWriter.add_scalers` method based on the contents of `values`.
        """
        self.writer.log(values, step=step)
        logger.debug("Successfully logged to swanlab")

    @on_main_process
    def log_images(self, values: dict, step: Optional[int], **kwargs):
        """
        Logs `images` to the current run.

        Args:
            values (Dictionary `str` to `List` of `np.ndarray` or `PIL.Image`):
                Values to be logged as key-value pairs. The values need to have type `List` of `np.ndarray` or
            step (`int`, *optional*):
                The run step. If included, the log will be affiliated with this step.
            kwargs:
                Additional key word arguments passed along to the `SummaryWriter.add_image` method.
        """

        for k, v in values.items():
            self.writer.log(k, swanlab.Image(v), step=step)

        logger.debug("Successfully logged images to swanlab")

    # @on_main_process
    # def log_table(
    #     self,
    #     table_name: str,
    #     columns: List[str] = None,
    #     data: List[List[Any]] = None,
    #     dataframe: Any = None,
    #     step: Optional[int] = None,
    #     **kwargs,
    # ):
    #     """
    #     Log a Table containing any object type (text, image, audio, video, molecule, html, etc). Can be defined either
    #     with `columns` and `data` or with `dataframe`.

    #     Args:
    #         table_name (`str`):
    #             The name to give to the logged table on the swanlab workspace
    #         columns (List of `str`'s *optional*):
    #             The name of the columns on the table
    #         data (List of List of Any data type *optional*):
    #             The data to be logged in the table
    #         dataframe (Any data type *optional*):
    #             The data to be logged in the table
    #         step (`int`, *optional*):
    #             The run step. If included, the log will be affiliated with this step.
    #     """

    #     values = {table_name: swanlab.Table(columns=columns, data=data, dataframe=dataframe)}
    #     self.log(values, step=step, **kwargs)

    @on_main_process
    def finish(self):
        """
        Closes `swanlab` writer
        """
        self.writer.finish()
        logger.debug("swanlab run closed")



import os
import sys
from datetime import datetime

from swankit.env import create_time
from swankit.log import FONT

MAX_UPLOAD_LEN = 500


class SwanWriterProxy:
    """
    标准输出流拦截代理
    """

    def __init__(self):
        self.epoch = 0
        self.write_callback = None
        self.file = None
        """
        当前正在写入的文件句柄
        """
        self.write_handler = None
        """
        标准输出流原始句柄
        """
        self.__buffer = ""
        """
        上传到云端的缓冲区
        """
        self.console_folder = None
        """
        保存控制台输出文件夹路径
        """
        self.now = None
        """
        当前文件名称（不包含后缀）
        """

    @property
    def can_callback(self) -> bool:
        return self.write_callback is not None

    def set_write_callback(self, func):
        # 封装一层func，加入epoch处理逻辑
        def _func(message):
            self.epoch += 1
            func({"message": message, "create_time": create_time(), "epoch": self.epoch})

        # 封装第二层，加入message处理逻辑以及是否调用逻辑
        def _(message):
            if self.can_callback:
                # 上传到云端
                messages = message.split("\n")
                # 如果长度为1，说明没有换行符
                if len(messages) == 1:
                    self.__buffer = self.__buffer + messages[0]
                # 如果长度大于2，说明其中包含多个换行符
                elif len(messages) > 1:
                    _func(self.__buffer + messages[0])
                    self.__buffer = messages[-1]
                    for m in messages[1:-1]:
                        _func(m)

        self.write_callback = _

    def init(self, path):
        self.console_folder = path
        # path 是否存在
        if not os.path.exists(path):
            os.makedirs(path)
        # 日志文件路径
        self.now = datetime.now().strftime("%Y-%m-%d")
        console_path = os.path.join(path, f"{self.now}.log")
        # 拿到日志文件句柄
        self.file = open(console_path, "a", encoding="utf-8")
        # 封装sys.stdout
        self.write_handler = sys.stdout.write

        def _(message):
            try:
                self.write_handler and self.write_handler(message)
            except UnicodeEncodeError:
                # 遇到编码问题，直接pass，此时表现为终端不输出
                pass
            except ValueError as e:
                if "I/O operation on closed file" in str(e):
                    # 遇到文件已关闭问题，直接pass，此时表现为终端不输出
                    pass
            # 限制上传长度
            message = FONT.clear(message)[:MAX_UPLOAD_LEN]
            self.write_callback and self.write_callback(message)

            # 检查文件分片
            now = datetime.now().strftime("%Y-%m-%d")
            if now != self.now:
                self.now = now
                if hasattr(self, "console") and not self.file.closed:
                    self.file.close()
                self.file = open(os.path.join(self.console_folder, self.now + ".log"), "a", encoding="utf-8")

            self.file.write(message)
            self.file.flush()

        sys.stdout.write = _

    def reset(self):
        sys.stdout.write = self.write_handler
        self.file and self.file.close()
        self.file = None
        self.write_callback = None


class SwanConsoler:
    def __init__(self):
        """
        控制台输出重定向器
        """
        self.writer = SwanWriterProxy()
        self.__installed = False

    @property
    def installed(self):
        return self.__installed

    def uninstall(self):
        """重置输出为原本的样子"""
        self.writer.reset()
        self.__installed = False

    def install(self, console_dir):
        """"""
        self.writer.init(console_dir)
        self.__installed = True

    @property
    def write_callback(self):
        return self.writer.write_callback

    def set_write_callback(self, func):
        self.writer.set_write_callback(func)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
r"""
@DATE: 2024-01-01 18:00:04
@File: swanlab/data/sdk.py
@IDE: vscode
@Description:
    在此处封装swanlab在日志记录模式下的各种接口
"""
import os
from typing import Optional, Union, Dict, Tuple, Literal

from swanboard import SwanBoardCallback
from swankit.env import SwanLabMode
from swankit.log import FONT

from swanlab.api import code_login, terminal_login
from swanlab.env import SwanLabEnv, is_interactive
from swanlab.log import swanlog
from .callback_cloud import CloudRunCallback
from .callback_local import LocalRunCallback
from .formatter import check_load_json_yaml, check_proj_name_format
from .modules import DataType
from .run import (
    SwanLabRunState,
    SwanLabRun,
    register,
    get_run,
)
from .run.helper import SwanLabRunOperator
from ..error import KeyFileError
from ..package import get_key, get_host_web, HostFormatter


def _check_proj_name(name: str) -> str:
    """检查项目名称是否合法，如果不合法则抛出ValueError异常
    项目名称必须是一个非空字符串，长度不能超过255个字符

    Parameters
    ----------
    name : str
        待检查的项目名称

    Returns
    -------
    str
        返回项目名称

    Raises
    ------
    ValueError
        项目名称不合法
    """
    _name = check_proj_name_format(name)
    if len(name) != len(_name):
        swanlog.warning(f"project name is too long, auto cut to {_name}")
    return _name


def login(api_key: str = None, host: str = None, web_host: str = None):
    """
    Login to SwanLab Cloud. If you already have logged in, you can use this function to relogin.
    Every time you call this function, the previous login information will be overwritten.

    [Note that] this function should be called before `init`.

    :param api_key: str, optional
        authentication key, if not provided, the key will be read from the key file.
    :param host: str, optional
        api host, if not provided, the default host will be used.
    :param web_host: str, optional
        web host, if not provided, the default host will be used.
    :return: LoginInfo
    """
    if SwanLabRun.is_started():
        raise RuntimeError("You must call swanlab.login() before using init()")
    # 检查host是否合法，并格式化，注入到环境变量中
    HostFormatter(host, web_host)()
    # 登录，初始化http对象
    CloudRunCallback.login_info = code_login(api_key) if api_key else CloudRunCallback.create_login_info()


MODES = Literal["disabled", "cloud", "local"]


def init(
    project: str = None,
    workspace: str = None,
    experiment_name: str = None,
    description: str = None,
    config: Union[dict, str] = None,
    logdir: str = None,
    mode: MODES = None,
    load: str = None,
    public: bool = None,
    **kwargs,
) -> SwanLabRun:
    """
    Start a new run to track and log. Once you have called this function, you can use 'swanlab.log' to log data to
    the current run. Meanwhile, you can use 'swanlab.finish' to finish the current run and close the current
    experiment. After calling this function, SwanLab will begin to record the console output of the current process,
    and register a callback function to the exit function.

    Parameters
    ----------
    project : str, optional
        The project name of the current experiment, the default is None,
        which means the current project name is the same as the current working directory.
    workspace : str, optional
        Where the current project is located, it can be an organization or a user (currently only supports yourself).
        The default is None, which means the current entity is the same as the current user.
    experiment_name : str, optional
        The experiment name you currently have open. If this parameter is not provided,
        SwanLab will generate one for you by default.
    description : str, optional
        The experiment description you currently have open,
        used for a more detailed introduction or labeling of the current experiment.
        If you do not provide this parameter, you can modify it later in the web interface.
    config : Union[dict, str], optional
        If you provide as a dict, it will be used as the configuration of the current experiment.
        If you provide as a string, SwanLab will read the configuration from the file.
        And the configuration file must be in the format of `json` or `yaml`.
        Anyway, you can modify the configuration later after this function is called.
    logdir : str, optional
        The folder will store all the log information generated during the execution of SwanLab.
        If the parameter is None,
        SwanLab will generate a folder named "swanlog" in the same path as the code execution to store the data.
        If you want to visualize the generated log files,
        simply run the command `swanlab watch` in the same path where the code is executed
        (without entering the "swanlog" folder).
        You can also specify your own folder, but you must ensure that the folder exists and preferably does not contain
        anything other than data generated by Swanlab.
        In this case, if you want to view the logs,
        you must use something like `swanlab watch -l ./your_specified_folder` to specify the folder path.
    mode : str, optional
        Allowed values are 'cloud', 'cloud-only', 'local', 'disabled'.
        If the value is 'cloud', the data will be uploaded to the cloud and the local log will be saved.
        If the value is 'cloud-only', the data will only be uploaded to the cloud and the local log will not be saved.
        If the value is 'local', the data will only be saved locally and will not be uploaded to the cloud.
        If the value is 'disabled', the data will not be saved or uploaded, just parsing the data.
    load : str, optional
        If you pass this parameter,SwanLab will search for the configuration file you specified
        (which must be in JSON or YAML format)
        and automatically fill in some explicit parameters of this function for you
        (excluding parameters in `**kwargs` and the parameters if they are None).
        In terms of priority, if the parameters passed to init are `None`,
        SwanLab will attempt to replace them from the configuration file you provided;
        otherwise, it will use the parameters you passed as the definitive ones.
    public : bool, optional
        Whether the project can be seen by anyone, the default is None, which means the project is private.
        Only available in cloud mode while the first time you create the project.
    """
    if SwanLabRun.is_started():
        swanlog.warning("You have already initialized a run, the init function will be ignored")
        return get_run()
    # ---------------------------------- 一些变量、格式检查 ----------------------------------

    # for https://github.com/SwanHubX/SwanLab/issues/809
    if experiment_name is None and kwargs.get("name", None) is not None:
        experiment_name = kwargs.get("name")
    if description is None and kwargs.get("notes", None) is not None:
        description = kwargs.get("notes")

    # 从文件中加载数据
    if load:
        load_data = check_load_json_yaml(load, load)
        experiment_name = _load_data(load_data, "experiment_name", experiment_name)
        description = _load_data(load_data, "description", description)
        config = _load_data(load_data, "config", config)
        logdir = _load_data(load_data, "logdir", logdir)
        mode = _load_data(load_data, "mode", mode)
        project = _load_data(load_data, "project", project)
        workspace = _load_data(load_data, "workspace", workspace)
        public = _load_data(load_data, "private", public)
    project = _check_proj_name(project if project else os.path.basename(os.getcwd()))  # 默认实验名称为当前目录名
    # ---------------------------------- 启动操作员 ----------------------------------
    operator, c = _create_operator(mode, public)
    exp_num = SwanLabRunOperator.parse_return(
        operator.on_init(project, workspace, logdir=logdir),
        key=c.__str__() if c else None,
    )
    # 初始化confi参数
    config = _init_config(config)
    # ---------------------------------- 实例化实验 ----------------------------------
    # 注册实验
    run = register(
        project_name=project,
        experiment_name=experiment_name,
        description=description,
        run_config=config,
        log_level=kwargs.get("log_level", "info"),
        exp_num=exp_num,
        operator=operator,
    )
    return run


def should_call_after_init(text):
    """
    装饰器，限制必须在实验初始化后调用
    """

    def decorator(func):
        def wrapper(*args, **kwargs):
            if not SwanLabRun.is_started():
                raise RuntimeError(text)
            return func(*args, **kwargs)

        return wrapper

    return decorator


@should_call_after_init("You must call swanlab.init() before using log()")
def log(data: Dict[str, DataType], step: int = None, print_to_console: bool = False):
    """
    Log a row of data to the current run.
    We recommend that you log data by SwanLabRun.log() method, but you can also use this function to log data.

    Parameters
    ----------
    data : Dict[str, DataType]
        Data must be a dict.
        The key must be a string with 0-9, a-z, A-Z, " ", "_", "-", "/".
        The value must be a `float`, `float convertible object`, `int` or `swanlab.data.BaseType`.
    step : int, optional
        The step number of the current data, if not provided, it will be automatically incremented.
        If step is duplicated, the data will be ignored.
    print_to_console : bool, optional
        Whether to print the data to the console, the default is False.
    """
    run = get_run()
    ll = run.log(data, step)
    print_to_console and print(ll)
    return ll


@should_call_after_init("You must call swanlab.init() before using finish()")
def finish(state: SwanLabRunState = SwanLabRunState.SUCCESS, error=None):
    """
    Finish the current run and close the current experiment
    Normally, swanlab will run this function automatically,
    but you can also execute it manually and mark the experiment as 'completed'.
    Once the experiment is marked as 'completed', no more data can be logged to the experiment by 'swanlab.log'.
    If you mark the experiment as 'CRASHED' manually, `error` must be provided.
    """
    run = get_run()
    if not run.running:
        return swanlog.error("After experiment is finished, you can't call finish() again.")
    run.finish(state, error)


def _init_mode(mode: str = None):
    """
    初始化mode参数
    从环境变量中提取默认的mode参数，如果传入的mode参数不为None，则使用环境变量中的mode参数，否则使用传入的mode参数
    传入的mode必须为SwanLabMode枚举中的一个值，否则报错ValueError
    如果环境变量和传入的mode参数都为None，则默认为cloud

    从环境变量中提取mode参数以后，还有一步让用户选择运行模式的交互，详见issue： https://github.com/SwanHubX/SwanLab/issues/632

    :param mode: str, optional
        传入的mode参数
    :return: str mode
    :raise ValueError: mode参数不合法
    """
    allowed = [m.value for m in SwanLabMode]
    mode_key = SwanLabEnv.MODE.value
    mode_value = os.environ.get(mode_key)
    if mode_value is not None and mode is not None:
        swanlog.warning(f"The environment variable {mode_key} will be overwritten by the parameter mode")
    mode = mode_value if mode is None else mode
    if mode is not None and mode not in allowed:
        raise ValueError(f"`mode` must be one of {allowed}, but got {mode}")
    mode = "cloud" if mode is None else mode
    # 如果mode为cloud，且没找到 api key或者未登录，则提示用户输入
    try:
        get_key()
        no_api_key = False
    except KeyFileError:
        no_api_key = True
    login_info = None
    # 三选一只允许登录官方的host，除非在此之前手动设置了环境变量
    # 详见 https://github.com/SwanHubX/SwanLab/issues/792#issuecomment-2603959483
    if mode == "cloud" and no_api_key:
        # 判断当前进程是否在交互模式下
        if is_interactive():
            swanlog.info(
                f"Using SwanLab to track your experiments. Please refer to {FONT.yellow('https://docs.swanlab.cn')} for more information."
            )
            swanlog.info("(1) Create a SwanLab account.")
            swanlog.info("(2) Use an existing SwanLab account.")
            swanlog.info("(3) Don't visualize my results.")

            web_host = get_host_web()
            # 交互选择
            tip = FONT.swanlab("Enter your choice: ")
            code = input(tip)
            while code not in ["1", "2", "3"]:
                swanlog.warning("Invalid choice, please enter again.")
                code = input(tip)
            if code == "3":
                mode = "local"
            elif code == "2":
                swanlog.info("You chose 'Use an existing swanlab account'")
                swanlog.info("Logging into " + FONT.yellow(web_host))
                login_info = terminal_login()
            elif code == "1":
                swanlog.info("You chose 'Create a swanlab account'")
                swanlog.info("Create a SwanLab account here: " + FONT.yellow(web_host + "/login"))
                login_info = terminal_login()
            else:
                raise ValueError("Invalid choice")
        # 如果不在就不管

    os.environ[mode_key] = mode
    return mode, login_info


def _init_config(config: Union[dict, str]):
    """初始化传入的config参数"""
    if isinstance(config, str):
        swanlog.info("The parameter config is loaded from the configuration file: {}".format(config))
        return check_load_json_yaml(config, "config")

    return config


def _load_data(load_data: dict, key: str, value):
    """
    从load_data中加载数据，如果value不是None，则直接返回value，如果为None，则返回load_data中的key
    """
    if value is not None:
        return value
    d = load_data.get(key, None)
    return d


def _create_operator(mode: str, public: bool) -> Tuple[SwanLabRunOperator, Optional[CloudRunCallback]]:
    """
    创建SwanLabRunOperator实例
    如果mode为disabled，则返回一个空的SwanLabRunOperator实例和None

    :param mode: 运行模式
    :param public: 是否公开
    :return: SwanLabRunOperator, CloudRunCallback
    """
    mode, login_info = _init_mode(mode)
    CloudRunCallback.login_info = login_info

    if mode == SwanLabMode.DISABLED.value:
        swanlog.warning("SwanLab run disabled, the data will not be saved or uploaded.")
        return SwanLabRunOperator(), None
    c = CloudRunCallback(public) if mode == SwanLabMode.CLOUD.value else LocalRunCallback()
    callbacks = [c, SwanBoardCallback()]
    return SwanLabRunOperator(callbacks), c



#!/usr/bin/env python
# -*- coding: utf-8 -*-
r"""
@DATE: 2024-03-19 15:30:24
@File: swanlab/auth/login.py
@IDE: vscode
@Description:
    用户登录接口，输入用户的apikey，保存用户token到本地
    进行一些交互定义和数据请求
"""
import getpass
import os
import sys

import requests
from swankit.env import is_windows
from swankit.log import FONT

from swanlab.api.info import LoginInfo
from swanlab.env import in_jupyter, SwanLabEnv
from swanlab.error import ValidationError, APIKeyFormatError
from swanlab.log import swanlog
from swanlab.package import get_setting_url, get_host_api, get_host_web


def login_request(api_key: str, api_host: str, timeout: int = 20) -> requests.Response:
    """用户登录，请求后端接口完成验证"""
    resp = requests.post(url=f"{api_host}/login/api_key", headers={"authorization": api_key}, timeout=timeout)
    return resp


def login_by_key(api_key: str, timeout: int = 20, save: bool = True) -> LoginInfo:
    """用户登录，异步调用接口完成验证
    返回后端内容(dict)，如果后端请求失败，返回None

    Parameters
    ----------
    api_key : str
        用户api_key
    timeout : int, optional
        请求认证的超时时间，单位秒
    save : bool, optional
        是否保存到本地token文件
    """
    api_host, web_host = get_host_api(), get_host_web()
    resp = login_request(api_key, api_host, timeout)
    # api key写入token文件
    login_info = LoginInfo(resp, api_key, api_host, web_host)
    save and not login_info.is_fail and login_info.save()
    return login_info


def input_api_key(
    tip: str = "Paste an API key from your profile and hit enter, or press 'CTRL + C' to quit",
    again: bool = False,
) -> str:
    """让用户输入apikey
    此时有两条消息，第一条消息为固定格式，第二条消息

    Parameters
    ----------
    tip : str
        提示信息
    again : bool, optional
        是否是重新输入api_key，如果是，不显示额外的提示信息
    """
    _t = sys.excepthook
    sys.excepthook = _abort_tip
    if not again:
        swanlog.info("You can find your API key at: " + FONT.yellow(get_setting_url()))
    # windows 额外打印提示信息
    if is_windows():
        tip += (
            '\nOn Windows, '
            f'use {FONT.yellow("Ctrl + Shift + V")} or {FONT.yellow("right-click")} '
            f'to paste the API key'
        )
    tip += ': '
    tip = FONT.swanlab(tip)
    ij = in_jupyter()
    ij and print(tip)
    key = getpass.getpass("" if ij else tip)
    sys.excepthook = _t
    return key


def code_login(api_key: str) -> LoginInfo:
    """
    代码内登录，此时会覆盖本地token文件（非task模式下）
    :param api_key: 用户的api_key
    :return: 登录信息
    :raise ValidationError: 登录失败
    :raise APIKeyFormatError: api_key格式错误
    """
    APIKeyFormatError.check(api_key)
    tip = "Waiting for the swanlab cloud response."
    save_key = os.environ.get(SwanLabEnv.RUNTIME.value) != 'task'
    login_info: LoginInfo = FONT.loading(tip, login_by_key, args=(api_key, 20, save_key), interval=0.5)
    if login_info.is_fail:
        raise ValidationError("Login failed: " + str(login_info))
    return login_info


def terminal_login(api_key: str = None) -> LoginInfo:
    """
    终端登录，此时直接覆盖本地token文件，但是新增交互，让用户输入api_key
    运行此函数，如果是认证失败的错误，重新要求用户输入api_key
    本地文件上层文件夹不保证存在，需要上层函数保证
    """
    # 1. api_key存在，跳过输入环节，直接请求登录接口，这与代码内swanlab.login方法一致
    # 2. api_key为None，提示用户输入
    is_input_key = api_key is None
    api_key = input_api_key() if api_key is None else api_key
    # 3. api_key校验失败且是输入的api_key，重新输入api_key，否则raise
    # 4. 登录失败且是输入的api_key，提示重新输入api_key，否则raise

    def login_again(error: Exception):
        swanlog.error(error)
        return input_api_key("Please try again, or press 'CTRL-C' to quit", True)

    while True:
        try:
            return code_login(api_key)
        # 登录失败且是输入的api_key，重新输入api_key
        except (APIKeyFormatError, ValidationError) as e:
            if not is_input_key:
                raise e
            api_key = login_again(e)


def _abort_tip(tp, _, __):
    """处理用户在input_api_key输入时按下CTRL+C的情况"""
    if tp == KeyboardInterrupt:
        sys.exit(0)
    # 如果不是CTRL+C，交给默认的异常处理



#!/usr/bin/env python
# -*- coding: utf-8 -*-
r"""
@DATE: 2024-01-01 15:58:58
@File: swanlab/data/run/main.py
@IDE: vscode
@Description:
    在此处定义SwanLabRun类并导出
"""
import random
from datetime import datetime
from typing import Callable, Optional, Dict

from swankit.core import SwanLabSharedSettings

from swanlab.data.modules import MediaType, DataWrapper, FloatConvertible, Line
from swanlab.env import get_mode, get_swanlog_dir
from swanlab.log import swanlog
from swanlab.package import get_package_version
from . import namer as N
from .config import SwanLabConfig
from .exp import SwanLabExp
from .helper import SwanLabRunOperator, RuntimeInfo, SwanLabRunState, MonitorCron, check_log_level
from .metadata import get_requirements, get_metadata
from .public import SwanLabPublicConfig
from ..formatter import check_key_format, check_exp_name_format, check_desc_format

MAX_LIST_LENGTH = 108


class SwanLabRun:
    """
    The SwanLabRun class is used for logging during a single experiment.
    There should be only one instance of the SwanLabRun class for each experiment.
    """

    def __init__(
        self,
        project_name: str = None,
        experiment_name: str = None,
        description: str = None,
        run_config=None,
        log_level: str = None,
        exp_num: int = None,
        operator: SwanLabRunOperator = SwanLabRunOperator(),
    ):
        """
        Initializing the SwanLabRun class involves configuring the settings and initiating other logging processes.

        Parameters
        ----------
        project_name : str, optional
            项目名称，目前单纯做一个记录
        experiment_name : str, optional
            实验名称，实验名称应该唯一，由0-9，a-z，A-Z，" ","_","-","/"组成
            如果不提供此参数(为None)，SwanLab将自动生成一个实验名称
        description : str, optional
            实验描述，用于对当前实验进行更详细的介绍或标注
            如果不提供此参数(为None)，可以在web界面中进行修改,这意味着必须在此改为空字符串""
        run_config : Any, optional
            实验参数配置，可以在web界面中显示，如学习率、batch size等
            不需要做任何限制，但必须是字典类型，可被json序列化，否则会报错
        log_level : str, optional
            当前实验的日志等级，默认为 'info'，可以从 'debug' 、'info'、'warning'、'error'、'critical' 中选择
            不区分大小写，如果不提供此参数(为None)，则默认为 'info'
            如果提供的日志等级不在上述范围内，默认改为info
        exp_num : int, optional
            历史实验总数，用于云端颜色与本地颜色的对应
        operator : SwanLabRunOperator, optional
            实验操作员，用于批量处理回调函数的调用，如果不提供此参数(为None)，则会自动生成一个实例
        """
        if self.is_started():
            raise RuntimeError("SwanLabRun has been initialized")

        # ---------------------------------- 初始化类内参数 ----------------------------------
        self.__project_name = project_name
        # 生成一个唯一的id，随机生成一个8位的16进制字符串，小写
        _id = hex(random.randint(0, 2**32 - 1))[2:].zfill(8)
        timestamp = datetime.now().strftime("%Y%m%d_%H%M%S")
        self.__run_id = "run-{}-{}".format(timestamp, _id)
        # 操作员初始化
        self.__operator = SwanLabRunOperator() if operator is None else operator
        self.__settings = SwanLabSharedSettings(
            logdir=get_swanlog_dir(),
            run_id=self.__run_id,
            should_save=not self.__operator.disabled,
            version=get_package_version(),
        )
        self.__mode = get_mode()
        self.__public = SwanLabPublicConfig(self.__project_name, self.__settings)
        self.__operator.before_run(self.__settings)
        # ---------------------------------- 初始化日志记录器 ----------------------------------
        swanlog.level = check_log_level(log_level)
        # ---------------------------------- 初始化配置 ----------------------------------
        # 如果config是以下几个类别之一，则抛出异常
        if isinstance(run_config, (int, float, str, bool, list, tuple, set)):
            raise TypeError(
                f"config: {run_config} (type: {type(run_config)}) is not a json serialized dict "
                f"(Support type is dict, MutableMapping, omegaconf.DictConfig, Argparse.Namespace), please check it"
            )
        global config
        config.update(run_config)
        setattr(config, "_SwanLabConfig__on_setter", self.__operator.on_runtime_info_update)
        self.__config = config
        # ---------------------------------- 注册实验 ----------------------------------
        self.__exp: SwanLabExp = self.__register_exp(experiment_name, description, num=exp_num)
        # 实验状态标记，如果status不为0，则无法再次调用log方法
        self.__state = SwanLabRunState.RUNNING

        # 动态定义一个方法，用于修改实验状态
        def _(state: SwanLabRunState):
            self.__state = state

        global _change_run_state, run
        _change_run_state = _
        run = self

        # ---------------------------------- 初始化完成 ----------------------------------
        self.__operator.on_run()
        # 执行__save，必须在on_run之后，因为on_run之前部分的信息还没完全初始化
        getattr(config, "_SwanLabConfig__save")()
        metadata, self.monitor_funcs = get_metadata(self.__settings.log_dir)
        # 系统信息采集
        self.__operator.on_runtime_info_update(
            RuntimeInfo(
                requirements=get_requirements(),
                metadata=metadata,
            )
        )
        # # 延时采集conda信息，因为conda信息的导出需要的时间会比较多
        # threading.Timer(
        #     5,
        #     lambda: self.__operator.on_runtime_info_update(RuntimeInfo(conda=get_conda())),
        # ).start()
        # 定时采集系统信息，目前仅cloud模式支持
        self.monitor_cron = None
        if self.mode == "cloud":
            if self.monitor_funcs is not None and len(self.monitor_funcs) != 0:
                swanlog.debug("Monitor on.")
                self.monitor_cron = MonitorCron(self.__get_monitor_func())
            else:
                swanlog.debug("Monitor off because of no monitor func.")
        else:
            swanlog.debug("Monitor off.")

    def __get_monitor_func(self):
        """
        获取监控函数
        """
        if self.monitor_funcs is None or len(self.monitor_funcs) == 0:
            return None

        def monitor_func():
            monitor_info_list = [f() for f in self.monitor_funcs]
            # 剔除其中为None的数据
            for monitor_info in monitor_info_list:
                if monitor_info is None:
                    swanlog.debug("Hardware info is empty. Skip it.")
                    continue
                for info in monitor_info:
                    key, name, value, cfg = (
                        info['key'],
                        info['name'],
                        info['value'],
                        info['config'],
                    )
                    v = DataWrapper(key, [Line(value)], reference="TIME")
                    self.__exp.add(
                        data=v,
                        key=key,
                        name=name,
                        column_config=cfg,
                        column_class="SYSTEM",
                        section_type="SYSTEM",
                    )

        return monitor_func

    def __register_exp(
        self,
        experiment_name: str = None,
        description: str = None,
        num: int = None,
    ) -> SwanLabExp:
        """
        注册实验，将实验配置写入数据库中，完成实验配置的初始化
        """
        if experiment_name:
            e = check_exp_name_format(experiment_name)
            if experiment_name != e:
                swanlog.warning("The experiment name has been truncated automatically.")
                experiment_name = e
        if description:
            d = check_desc_format(description)
            if description != d:
                swanlog.warning("The description has been truncated automatically.")
                description = d
        experiment_name = N.generate_name(num) if experiment_name is None else experiment_name
        description = "" if description is None else description
        colors = N.generate_colors(num)
        self.__operator.before_init_experiment(self.__run_id, experiment_name, description, num, colors)
        self.__settings.exp_name = experiment_name
        self.__settings.exp_colors = colors
        self.__settings.description = description
        return SwanLabExp(self.__settings, operator=self.__operator)

    def __cleanup(self, error: str = None):
        """
        停止部分功能，内部清理时调用
        """
        if self.monitor_cron is not None:
            self.monitor_cron.cancel()
        self.__operator.on_stop(error)

    def __str__(self) -> str:
        """此类的字符串表示"""
        return self.__run_id

    @property
    def public(self):
        return self.__public

    @property
    def mode(self) -> str:
        return self.__mode

    @property
    def state(self) -> SwanLabRunState:
        return self.__state

    @classmethod
    def get_state(cls) -> SwanLabRunState:
        """
        获取当前实验状态
        """
        global run
        return run.state if run is not None else SwanLabRunState.NOT_STARTED

    @staticmethod
    def is_started() -> bool:
        """
        If the experiment has been initialized, return True, otherwise return False.
        """
        return get_run() is not None

    @property
    def crashed(self) -> bool:
        """
        If the experiment is marked as 'CRASHED', return True, otherwise return False.
        """
        return self.__state == SwanLabRunState.CRASHED

    @property
    def success(self) -> bool:
        """
        If the experiment is marked as 'SUCCESS', return True, otherwise return False.
        """
        return self.__state == SwanLabRunState.SUCCESS

    @property
    def running(self) -> bool:
        """
        If the experiment is marked as 'RUNNING', return True, otherwise return False.
        """
        return self.__state == SwanLabRunState.RUNNING

    @staticmethod
    def finish(state: SwanLabRunState = SwanLabRunState.SUCCESS, error=None):
        """
        Finish the current run and close the current experiment
        Normally, swanlab will run this function automatically,
        but you can also execute it manually and mark the experiment as 'completed'.
        Once the experiment is marked as 'completed', no more data can be logged to the experiment by 'swanlab.log'.
        After calling this function, you can re-run `swanlab.init` to start a new experiment.

        :param state: The state of the experiment, it can be 'SUCCESS', 'CRASHED' or 'RUNNING'.
        :param error: The error message when the experiment is marked as 'CRASHED'. If not 'CRASHED', it should be None.
        """
        global run, config
        # 分为几步
        # 1. 设置数据库实验状态为对应状态
        # 2. 判断是否为云端同步，如果是则开始关闭线程池和同步状态
        # 3. 清空run和config对象，run改为局部变量_run，新建一个config对象，原本的config对象内容转移到新的config对象，全局config被清空
        # 4. 返回_run
        if run is None:
            raise RuntimeError("The run object is None, please call `swanlab.init` first.")
        if state == SwanLabRunState.CRASHED and error is None:
            raise ValueError("When the state is 'CRASHED', the error message cannot be None.")
        _set_run_state(state)
        error = error if state == SwanLabRunState.CRASHED else None
        # 退出回调
        getattr(run, "_SwanLabRun__cleanup")(error)
        try:
            swanlog.uninstall()
        except RuntimeError:
            # disabled 模式下没有install，所以会报错
            pass

        # ---------------------------------- 清空config和run ----------------------------------
        _config = SwanLabConfig(config)
        setattr(run, "_SwanLabRun__config", _config)
        config.clean()
        _run, run = run, None

        return _run

    @property
    def config(self) -> SwanLabConfig:
        """
        This property allows you to access the 'config' content passed through `init`,
        and allows you to modify it. The latest configuration after each modification
        will be synchronized to the corresponding path by Swanlab. If you have
        enabled the web service, you will notice the changes after refreshing.
        """
        return self.__config

    def __flatten_dict(self, d: dict, parent_key='', sep='.') -> dict:
        """Helper method to flatten nested dictionaries with dot notation"""
        items = []
        for k, v in d.items():
            new_key = f"{parent_key}{sep}{k}" if parent_key else k
            if isinstance(v, dict):
                items.extend(self.__flatten_dict(v, new_key, sep=sep).items())
            else:
                items.append((new_key, v))
        return dict(items)

    def log(self, data: dict, step: int = None):
        """
        Log a row of data to the current run. Unlike `swanlab.log`, this api will be called directly based on the
        SwanRun instance, removing the initialization process. Of course, after you call the success/fail method,
        this method will be banned from calling.

        Parameters
        ----------
        data : Dict[str, DataType]
            Data must be a dict.
            The key must be a string with 0-9, a-z, A-Z, " ", "_", "-", "/".
            The value must be a `float`, `float convertible object`, `int`, `swanlab.data.BaseType`, or a nested dict.
            For nested dicts, keys will be joined with dots (e.g., {'a': {'b': 1}} becomes {'a.b': 1}).
        step : int, optional
            The step number of the current data, if not provided, it will be automatically incremented.
            If step is duplicated, the data will be ignored.

        Raises
        ----------
        ValueError:
            Unsupported key names.
        """
        if self.__state != SwanLabRunState.RUNNING:
            raise RuntimeError("After experiment finished, you can no longer log data to the current experiment")
        self.__operator.on_log()

        if not isinstance(data, dict):
            return swanlog.error(
                "log data must be a dict, but got {}, SwanLab will ignore records it.".format(type(data))
            )
        # 检查step的类型
        if step is not None and (not isinstance(step, int) or step < 0):
            swanlog.error(
                "'step' must be an integer not less than zero, but got {}, SwanLab will automatically set step".format(
                    step
                )
            )
            step = None

        # 展平嵌套字典
        flattened_data = self.__flatten_dict(data)

        log_return = {}
        # 遍历data，记录data
        for k, v in flattened_data.items():
            _k = k
            k = check_key_format(k, auto_cut=True)
            if k != _k:
                # 超过255字符，截断
                swanlog.warning(f"Key {_k} is too long, cut to 255 characters.")
                if k in flattened_data.keys():
                    raise ValueError(f'tag: Not supported too long Key "{_k}" and auto cut failed')
            # ---------------------------------- 包装数据 ----------------------------------
            # 输入为可转换为float的数据类型
            if isinstance(v, (int, float, FloatConvertible)):
                v = DataWrapper(k, [Line(v)])
            # 为Line类型或者MediaType类型
            elif isinstance(v, (Line, MediaType)):
                v = DataWrapper(k, [v])
            # 为List[MediaType]或者List[Line]类型，且长度大于0，且所有元素类型相同
            elif (
                isinstance(v, list)
                and len(v) > 0
                and all([isinstance(i, (Line, MediaType)) for i in v])
                and all([i.__class__ == v[0].__class__ for i in v])
            ):
                if len(v) > MAX_LIST_LENGTH:
                    swanlog.warning(f"List length '{k}' is too long, cut to {MAX_LIST_LENGTH}.")
                    v = v[:MAX_LIST_LENGTH]
                v = DataWrapper(k, v)
            else:
                # 其余情况被当作是非法的数据类型，交给Line处理
                v = DataWrapper(k, [Line(v)])
            # 数据类型的检查将在创建chart配置的时候完成，因为数据类型错误并不会影响实验进行
            metric_info = self.__exp.add(key=k, data=v, step=step)
            log_return[metric_info.column_info.key] = metric_info

        return log_return


run: Optional["SwanLabRun"] = None
"""Global runtime instance. After the user calls finish(), run will be set to None."""
_change_run_state: Optional["Callable"] = None
"""
修改实验状态的函数，用于在实验状态改变时调用
"""

# 全局唯一的config对象，不应该重新赋值
config: Optional["SwanLabConfig"] = SwanLabConfig()
"""
Global config instance.
"""


def _set_run_state(state: SwanLabRunState):
    """
    设置实验状态，只能在run对象存在的情况下调用
    """
    if run is None:
        raise RuntimeError("The run object is None, please call swanlab.init first.")

    if state not in SwanLabRunState or state in [SwanLabRunState.NOT_STARTED, SwanLabRunState.RUNNING]:
        swanlog.warning("Invalid state when set, state must be in SwanLabRunState and not be RUNNING or NOT_STARTED")
        swanlog.warning("SwanLab will set state to `CRASHED`")
        state = SwanLabRunState.CRASHED
    # 设置state
    _change_run_state(state)


def get_run() -> Optional["SwanLabRun"]:
    """
    Get the current run object. If the experiment has not been initialized, return None.
    """
    global run
    return run


def get_config() -> Optional["SwanLabConfig"]:
    """
    Get the current config object.
    """
    global config
    return config


def get_url() -> Optional["str"]:
    """
    Get the url of the current experiment.
    NOTE: return None if the experiment has not been initialized or mode is not 'cloud'.
    """
    global run
    if run is None:
        return None
    return run.public.cloud.experiment_url


def get_project_url() -> Optional["str"]:
    """
    Get the url of the current project.
    NOTE: return None if the experiment has not been initialized or mode is not 'cloud'.
    """
    global run
    if run is None:
        return None
    return run.public.cloud.project_url



from .console import SwanConsoler
from swankit.log import SwanLabSharedLog


class SwanLog(SwanLabSharedLog):

    def __init__(self, name=__name__.lower(), level="info"):
        super().__init__(name=name, level=level)
        # 控制台监控记录器
        self.__consoler = SwanConsoler()
        self.__original_level = level
        self.__installed = False

    @property
    def installed(self):
        """
        判断是否已经install
        """
        return self.__installed

    def install(self, console_dir: str = None, log_level: str = None) -> "SwanLog":
        """
        初始化安装日志系统，同一实例在没有执行uninstall的情况下，不可重复安装
        功能是开启标准输出流拦截功能，并设置日志等级
        :param console_dir: 控制台日志文件路径文件夹，如果提供，则会将控制台日志记录到对应文件夹，否则不记录，需要保证文件夹存在
        :param log_level: 日志等级，可以是 "debug", "info", "warning", "error", 或 "critical"，默认为info

        :return: SwanLog实例

        :raises: RuntimeError: 已经安装过日志系统
        :raises: KeyError: 无效的日志级别
        :raises: FileNotFoundError: 控制台日志文件夹不存在
        """
        if self.installed:
            raise RuntimeError("SwanLog has been installed")
        # 设置日志等级
        if log_level is not None:
            self.level = log_level
        # 初始化控制台记录器
        if console_dir:
            self.debug("Init consoler to record console log")
            self.__consoler.install(console_dir)
        self.__installed = True
        return self

    def uninstall(self):
        """
        卸载日志系统，卸载后需要重新安装
        在设计上我们并不希望外界乱用这个函数，所以我们不提供外部调用（不在最外层的__all__中）
        此时将卸载标准输出流拦截功能，并重置日志等级为初始化时的等级
        """
        if not self.installed:
            raise RuntimeError("SwanLog has not been installed")
        self.debug("uninstall swanlog, reset consoler")
        self.level = self.__original_level
        if self.__consoler.installed:
            self.__consoler.uninstall()
        self.__installed = False

    @property
    def write_callback(self):
        return self.__consoler.write_callback

    def set_write_callback(self, func):
        self.__consoler.set_write_callback(func)

    @property
    def epoch(self):
        """
        获取当前日志的 epoch
        """
        return self.__consoler.writer.epoch

    @property
    def file(self):
        if self.__consoler.installed:
            return self.__consoler.writer.file
        else:
            return None



#!/usr/bin/env python
# -*- coding: utf-8 -*-
r"""
@DATE: 2024/5/31 12:47
@File: pytest_cos.py
@IDE: pycharm
@Description:
    测试http的api
    开发环境下存储凭证过期时间为3s
"""
import os
import nanoid
from swanlab.api.http import create_http, HTTP, CosClient
from swanlab.api.auth.login import login_by_key
from swanlab.data.modules import MediaBuffer
from tutils import API_KEY, TEMP_PATH, is_skip_cloud_test
from tutils.setup import UseMocker, UseSetupHttp
import pytest
import responses
from responses import registries

alphabet = "abcdefghijklmnopqrstuvwxyz"


def test_decode_response():
    with UseMocker() as mocker:
        mocker.post("/json", json={"test": "test"})
        mocker.post("/text", text="test")
        with UseSetupHttp() as http:
            data = http.post("/json")
            assert data == {"test": "test"}
            data = http.post("/text")
            assert data == "test"


@responses.activate(registry=registries.OrderedRegistry)
def test_retry():
    """
    测试重试机制
    """
    from swanlab.package import get_host_api
    url = get_host_api() + "/retry"
    rsp1 = responses.get(url, body="Error", status=500)
    rsp2 = responses.get(url, body="Error", status=500)
    rsp3 = responses.get(url, body="Error", status=500)
    rsp4 = responses.get(url, body="OK", status=200)
    with UseSetupHttp() as http:
        data = http.get("/retry")
        assert data == "OK"
        assert rsp1.call_count == 1
        assert rsp2.call_count == 1
        assert rsp3.call_count == 1
        assert rsp4.call_count == 1


@pytest.mark.skipif(is_skip_cloud_test, reason="skip cloud test")
class TestCosSuite:
    http: HTTP = None
    project_name = nanoid.generate(alphabet)
    experiment_name = nanoid.generate(alphabet)
    file_path = os.path.join(TEMP_PATH, nanoid.generate(alphabet))
    now_refresh_time = 1
    pre_refresh_time = CosClient.REFRESH_TIME

    @classmethod
    def setup_class(cls):
        CosClient.REFRESH_TIME = cls.now_refresh_time
        # 这里不测试保存token的功能
        login_info = login_by_key(API_KEY, save=False)
        cls.http = create_http(login_info)
        cls.http.mount_project(cls.project_name)
        cls.http.mount_exp(cls.experiment_name, ('#ffffff', '#ffffff'))
        # temp路径写一个文件上传
        with open(cls.file_path, "w") as f:
            f.write("test")

    @classmethod
    def teardown_class(cls):
        CosClient.REFRESH_TIME = cls.pre_refresh_time

    def test_cos_ok(self):
        assert self.http is not None
        assert self.http.cos is not None

    def test_cos_upload(self):
        # 新建一个文件对象
        buffer = MediaBuffer()
        buffer.write(b"test")
        buffer.file_name = "test"
        self.http.upload(buffer)
        # 为了开发方便，测试刷新功能关闭

        # # 开发版本设置的过期时间为3s，等待过期
        # time.sleep(3)
        # # 重新上传，测试刷新
        # assert self.http.cos.should_refresh is True
        # self.http.upload(buffer)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
r"""
@DATE: 2024/4/7 16:56
@File: __init__.py
@IDE: pycharm
@Description:
    上传相关接口
"""
from typing import List

from swanlab.error import ApiError
from swanlab.log import swanlog
from .model import ColumnModel, MediaModel, ScalarModel, FileModel
from ..http import get_http, sync_error_handler

house_url = '/house/metrics'


def create_data(metrics: List[dict], metrics_type: str) -> dict:
    """
    携带上传日志的指标信息
    """
    http = get_http()
    return {"projectId": http.proj_id, "experimentId": http.exp_id, "type": metrics_type, "metrics": metrics}


@sync_error_handler
def upload_logs(logs: List[dict], level: str = "INFO"):
    """
    上传日志信息
    :param logs: 日志列表
    :param level: 日志级别，'INFO', 'ERROR'，默认INFO
    """
    http = get_http()
    # 将logs解析为json格式
    metrics = [{"level": level, **x} for x in logs]
    data = create_data(metrics, "log")
    http.post(house_url, data)


@sync_error_handler
def upload_media_metrics(media_metrics: List[MediaModel]):
    """
    上传指标的媒体数据
    :param media_metrics: 媒体指标数据集合
    """
    http = get_http()
    buffers = []
    for media in media_metrics:
        media.buffers and buffers.extend(media.buffers)
    http.upload_files(buffers)
    # 上传指标信息
    http.post(house_url, create_data([x.to_dict() for x in media_metrics], MediaModel.type.value))


@sync_error_handler
def upload_scalar_metrics(scalar_metrics: List[ScalarModel]):
    """
    上传指标的标量数据
    """
    http = get_http()
    data = create_data([x.to_dict() for x in scalar_metrics], ScalarModel.type.value)
    http.post(house_url, data)


_valid_files = {
    'config.yaml': ['config', 'yaml'],
    'requirements.txt': ['requirements', 'txt'],
    'swanlab-metadata.json': ['metadata', 'json'],
}
"""
支持上传的文件列表，filename: key
"""


@sync_error_handler
def upload_files(files: List[FileModel]):
    """
    上传files文件夹中的内容
    :param files: 文件列表，内部为文件绝对路径
    """
    http = get_http()
    # 去重所有的FileModel，留下一个
    if len(files) == 0:
        return swanlog.warning("No files to upload.")
    file_model = FileModel.create(files)
    # 如果没有文件需要上传，直接返回
    if file_model.empty:
        return
    data = file_model.to_dict()
    http.put(f'/project/{http.groupname}/{http.projname}/runs/{http.exp_id}/profile', data)


@sync_error_handler
def upload_column(columns: List[ColumnModel]):
    """
    上传列信息，需要注意的是一次只能上传一个列，所以函数名不带s
    但是在设计上是聚合上传的，所以在此处需要进行拆分然后分别上传
    """
    http = get_http()
    url = f'/experiment/{http.exp_id}/column'
    # WARNING 这里不能使用并发请求，可见 https://github.com/SwanHubX/SwanLab-Server/issues/113
    for column in columns:
        try:
            http.post(url, column.to_dict())
        except ApiError as e:
            swanlog.error(f"Upload column {column.key} failed: {e.resp.status_code}")


__all__ = [
    "upload_logs",
    "upload_media_metrics",
    "upload_scalar_metrics",
    "upload_files",
    "upload_column",
    "ScalarModel",
    "MediaModel",
    "ColumnModel",
]



#!/usr/bin/env python
# -*- coding: utf-8 -*-
r"""
@DATE: 2023-12-15 17:33:56
@File: swanlab/log/__init__.py
@IDE: vscode
@Description:
    日志记录模块，在设计上swanlog作为一个独立的模块被使用，你可以在除了utils的任何地方使用它
"""
from typing import Optional
from .log import SwanLog

swanlog: Optional["SwanLog"] = SwanLog("swanlab")

install = swanlog.install

uninstall = swanlog.uninstall

__all__ = ["install", "uninstall", "swanlog"]



from .audio import Audio
from .image import Image
from .text import Text
from .line import Line, FloatConvertible
from typing import Union, List
from .wrapper import DataWrapper
from swankit.core.data import MediaType, MediaBuffer, BaseType

DataType = Union[int, float, FloatConvertible, BaseType, List[BaseType]]
ChartType = BaseType.Chart

__all__ = [
    "FloatConvertible",
    "DataWrapper",
    "MediaType",
    "Image",
    "Audio",
    "Text",
    "Line",
    "DataType",
    "ChartType",
    "MediaBuffer"
]



#!/usr/bin/env python
# -*- coding: utf-8 -*-
r"""
@DATE: 2024/6/19 16:46
@File: callback.py
@IDE: pycharm
@Description:
    回调函数注册抽象模块
"""
from typing import Optional
from swankit.core import SwanLabSharedSettings
from swankit.callback import SwanKitCallback
from swanlab.log import swanlog
from swankit.log import FONT
from swanlab.env import is_windows
from swanlab.package import get_package_version
import atexit
import sys
import os


class U:
    """
    工具函数类，隔离SwanLabRunCallback回调与其他工具函数
    """

    def __init__(self):
        self.settings: Optional[SwanLabSharedSettings] = None

    @staticmethod
    def fmt_windows_path(path: str) -> str:
        """这主要针对windows环境，输入的绝对路径可能不包含盘符，这里进行补充
        主要是用于打印效果
        如果不是windows环境，直接返回path，相当于没有调用这个函数

        Parameters
        ----------
        path : str
            待转换的路径

        Returns
        -------
        str
            增加了盘符的路径
        """
        if not is_windows():
            return path
        if not os.path.isabs(path):
            return path
        need_add = len(path) < 3 or path[1] != ":"
        # 处理反斜杠, 保证路径的正确性
        path = path.replace("/", "\\")
        if need_add:
            return os.path.join(os.getcwd()[:2], path)
        return path

    def _train_begin_print(self):
        """
        训练开始时的打印信息
        """
        swanlog.debug("SwanLab Runtime has initialized")
        swanlog.debug("SwanLab will take over all the print information of the terminal from now on")
        swanlog.info("Tracking run with swanlab version " + get_package_version())
        local_path = FONT.magenta(FONT.bold(self.fmt_windows_path(self.settings.run_dir)))
        swanlog.info("Run data will be saved locally in " + local_path)

    def _watch_tip_print(self):
        """
        watch命令提示打印
        """
        swanlog.info(
            "🌟 Run `"
            + FONT.bold("swanlab watch {}".format(self.fmt_windows_path(self.settings.swanlog_dir)))
            + "` to view SwanLab Experiment Dashboard locally"
        )

    def _train_finish_print(self):
        """
        打印结束信息
        """
        swanlog.info("Experiment {} has completed".format(FONT.yellow(self.settings.exp_name)))


class SwanLabRunCallback(SwanKitCallback, U):
    """
    SwanLabRunCallback，回调函数注册类，所有以`on_`和`before_`开头的函数都会在对应的时机被调用
    为了方便管理：
    1. `_`开头的函数为内部函数，不会被调用，且写在最开头
    2. 所有回调按照逻辑上的触发顺序排列
    3. 带有from_*后缀的回调函数代表调用者来自其他地方，比如config、operator等，这将通过settings对象传递
    4. 所有回调不要求全部实现，只需实现需要的回调即可
    """

    def _register_sys_callback(self):
        """
        注册系统回调，内部使用
        """
        sys.excepthook = self._except_handler
        atexit.register(self._clean_handler)

    def _unregister_sys_callback(self):
        """
        注销系统回调，内部使用
        """
        sys.excepthook = sys.__excepthook__
        atexit.unregister(self._clean_handler)

    def _clean_handler(self):
        """
        正常退出清理函数，此函数调用`run.finish`
        """
        pass

    def _except_handler(self, tp, val, tb):
        """
        异常退出清理函数
        """
        pass

    def __str__(self):
        raise NotImplementedError("Please implement this method")



#!/usr/bin/env python
# -*- coding: utf-8 -*-
r"""
@DATE: 2024/7/27 15:07
@File: setup.py
@IDE: pycharm
@Description:
    测试tutils/setup.py
"""
import pytest
import tutils.setup as SU


def test_mock_login_info():
    login_info = SU.mock_login_info()
    assert login_info.is_fail is False
    login_info = SU.mock_login_info(error_reason="Unauthorized")
    assert login_info.is_fail is True
    login_info = SU.mock_login_info(error_reason="Authorization Required")
    assert login_info.is_fail is True
    login_info = SU.mock_login_info(error_reason="Forbidden")
    assert login_info.is_fail is True
    login_info = SU.mock_login_info(error_reason="OK")
    assert login_info.is_fail is False


def test_use_setup_http():
    from swanlab.api import get_http
    with SU.UseSetupHttp() as http:
        assert http is not None
        assert get_http() is not None
    with pytest.raises(ValueError):
        get_http()


def test_use_mocker():
    with SU.UseMocker() as m:
        m.post("/tmp", text="mock")
        import requests
        from swanlab.package import get_host_api
        resp = requests.post(get_host_api() + "/tmp")
        assert resp.text == "mock"



import swanlab

def _extract_args(args, kwargs, param_names):
    """
    从args和kwargs中提取参数值的通用函数
    
    Args:
        args: 位置参数元组
        kwargs: 关键字参数字典
        param_names: 参数名称列表
    
    Returns:
        tuple: 按param_names顺序返回提取的参数值
    """
    values = []
    for i, name in enumerate(param_names):
        if len(args) > i:
            values.append(args[i])
        else:
            values.append(kwargs.get(name, None))
    return tuple(values)


def sync_wandb(mode:str="cloud", wandb_run:bool=True):
    """
    sync wandb with swanlab, 暂时不支持log非标量类型
    
    - mode: "cloud", "local" or "disabled". https://docs.swanlab.cn/api/py-init.html
    - wandb_run: 如果此参数设置为False，则不会将数据上传到wandb，等同于设置wandb.init(mode="offline")。
    
    usecase:
    ```python
    import wandb
    import random
    import swanlab

    swanlab.sync_wandb()
    # swanlab.init(project="sync_wandb")

    wandb.init(
        project="test",
        config={"a": 1, "b": 2},
        name="test",
    )

    epochs = 10
    offset = random.random() / 5
    for epoch in range(2, epochs):
        acc = 1 - 2 ** -epoch - random.random() / epoch - offset
        loss = 2 ** -epoch + random.random() / epoch + offset

        wandb.log({"acc": acc, "loss": loss})
    ```
    """
    try:
        import wandb
        from wandb import sdk as wandb_sdk
    except ImportError:
        raise ImportError("please install wandb first, command: `pip install wandb`")
    
    original_init = wandb.init
    original_log = wandb_sdk.wandb_run.Run.log
    original_finish = wandb_sdk.finish
    original_config_update = wandb_sdk.wandb_config.Config.update
    
    def patched_init(*args, **kwargs):
        entity, project, dir, id, name, notes, tags, config, config_exclude_keys = _extract_args(
            args, kwargs, ['entity', 'project', 'dir', 'id', 'name', 'notes', 'tags', 'config', 'config_exclude_keys']
        )
        
        if swanlab.data.get_run() is None:
            swanlab.init(
                project=project,
                experiment_name=name,
                description=notes,
                config=config,
                mode=mode)
        else:
            swanlab.config.update(config)
        
        if wandb_run is False:
            kwargs["mode"] = "offline"
            return original_init(*args, **kwargs)
        else:
            return original_init(*args, **kwargs)

    def patched_config_update(self, *args, **kwargs):
        d, _ = _extract_args(args, kwargs, ['d', 'allow_val_change'])
        
        if d is not None:
            swanlab.config.update(d)
        return original_config_update(self, *args, **kwargs)

    def patched_log(self, *args, **kwargs):
        data, step, commit, sync = _extract_args(args, kwargs, ['data', 'step', 'commit', 'sync'])
        
        if data is None:
            return original_log(self, *args, **kwargs)
        
        # 过滤掉非标量类型
        filtered_data = {}
        for key, value in data.items():
            if isinstance(value, (int, float, bool, str)):
                filtered_data[key] = value
        
        swanlab.log(data=filtered_data, step=step)
        
        return original_log(self, *args, **kwargs)
    
    def patched_finish(*args, **kwargs):
        swanlab.finish()
        return original_finish(*args, **kwargs)

    wandb.init = patched_init
    wandb_sdk.wandb_run.Run.log = patched_log
    wandb_sdk.wandb_run.Run.finish = patched_finish
    wandb_sdk.wandb_config.Config.update = patched_config_update


#!/usr/bin/env python
# -*- coding: utf-8 -*-
r"""
@DATE: 2024/4/8 19:47
@File: cos.py
@IDE: pycharm
@Description:
    tencent cos
"""
# noinspection PyPackageRequirements
from qcloud_cos import CosConfig
# noinspection PyPackageRequirements
from qcloud_cos import CosS3Client
# noinspection PyPackageRequirements
from qcloud_cos.cos_threadpool import SimpleThreadPool
from datetime import datetime, timedelta
from typing import List, Dict, Union
from swanlab.data.modules import MediaBuffer
from swanlab.log import swanlog


class CosClient:
    REFRESH_TIME = 60 * 60 * 1.5  # 1.5小时

    def __init__(self, data):
        """
        初始化cos客户端
        """
        self.__expired_time = datetime.fromtimestamp(data["expiredTime"])
        self.__prefix = data["prefix"]
        self.__bucket = data["bucket"]
        credentials = data["credentials"]
        config = CosConfig(
            Region=data["region"],
            SecretId=credentials['tmpSecretId'],
            SecretKey=credentials['tmpSecretKey'],
            Token=credentials['sessionToken'],
            Scheme='https'
        )
        self.__client = CosS3Client(config)

    def upload(self, buffer: MediaBuffer):
        """
        上传文件，需要注意的是file_path应该为unix风格而不是windows风格
        开头不能有/
        :param buffer: 本地文件的二进制数据
        """
        key = "{}/{}".format(self.__prefix, buffer.file_name)
        try:
            swanlog.debug("Uploading file: {}".format(key))
            self.__client.put_object(
                Bucket=self.__bucket,
                Key=key,
                Body=buffer.getvalue(),
                EnableMD5=False,
                # 一年
                CacheControl="max-age=31536000",
            )
        except Exception as e:
            swanlog.error("Upload error: {}".format(e))

    def upload_files(self, buffers: List[MediaBuffer]) -> Dict[str, Union[bool, List]]:
        """
        批量上传文件，keys和local_paths的长度应该相等
        :param buffers: 本地文件的二进制对象集合
        """
        pool = SimpleThreadPool()
        for buffer in buffers:
            self.upload(buffer)
        pool.wait_completion()
        result = pool.get_result()
        return result

    @property
    def should_refresh(self):
        # cos传递的是北京时间，需要添加8小时
        now = datetime.utcnow() + timedelta(hours=8)
        # 过期时间减去当前时间小于刷新时间，需要注意为负数的情况
        if self.__expired_time < now:
            return True
        return (self.__expired_time - now).seconds < self.REFRESH_TIME
